"""Support module for extra files with key-value information
related to Eclipse runs"""

import json
import logging
from pathlib import Path
from typing import Any, Dict, List, Union

import pandas as pd
import yaml

from ecl2df.eclfiles import EclFiles

logger = logging.getLogger(__name__)


def find_parameter_files(
    ecldeck_or_eclpath: Union[EclFiles, str, Path], filebase: str = "parameters"
) -> List[Path]:
    """Locate a default prioritized list of files to try to read as key-value

    File extensions .yml, .json and .txt are recognized and will be found in
    current dir, one directory up, and two directories up.

    Args:
        ecldeck_or_eclpath: Either an EclFiles object of
            an Eclipse output set (only the corresponding path will be used),
            or path to a file or directory, that will be used as a starting
            point for locating parameter files
        filebase: the base of filenames to look for.

    Return:
        Absolute paths to filenames. Empty list if nothing found
    """
    eclbasepath: Path
    fname: str
    if isinstance(ecldeck_or_eclpath, EclFiles):
        eclbasepath = Path(ecldeck_or_eclpath.get_path())
    elif isinstance(ecldeck_or_eclpath, (str, Path)):
        eclbasepath = Path(ecldeck_or_eclpath).parent.absolute()
    else:
        raise TypeError
    files_to_lookfor: List[str] = [
        filebase + ".json",
        filebase + ".yml",
        filebase + ".txt",
        filebase,
    ]
    paths_to_check: List[Path] = [Path("."), Path(".."), Path("..") / Path("..")]
    foundfiles = []
    for path in paths_to_check:
        for fname in files_to_lookfor:
            fullfname = eclbasepath / path / Path(fname)
            if fullfname.is_file():
                foundfiles.append(fullfname.resolve())
    return foundfiles


def load_parameterstxt(filename: Union[str, Path]) -> Dict[str, Any]:
    """Read parameters.txt into a dictionary

    Lines starting with a hash will be ignored.

    Args:
        filename: file containing one key-value pair pr. line,
            separated by whitespace
    """
    dframe = pd.read_csv(
        filename, comment="#", sep=r"\s", engine="python", names=["KEY", "VALUE"]
    )
    return dframe.set_index("KEY")["VALUE"].to_dict()


def load_all(
    filenames: Union[List[str], List[Path]], warnduplicates: bool = True
) -> Dict[str, Any]:
    """Reads a list of parameter filenames

    Dictionaries for all files will be merged into one.

    Keys must be unique over all filenames, if not
    only the first occurence will be used (based on
    the order of filenames).

    Args:
        filenames: Order matters.
        warnduplicates: If True (default), overlapping keys will be warned.
    """
    keyvalues: Dict[str, Any] = {}
    for fname in filenames:
        new_params = load(fname)
        if warnduplicates and keyvalues:
            duplicates = set(keyvalues.keys()).intersection(set(new_params.keys()))
            if duplicates:
                logger.debug("Duplicates keys %s", str(duplicates))
        new_params.update(keyvalues)
        keyvalues = new_params
    return keyvalues


def load(filename: Union[str, Path]) -> Dict[str, Any]:
    """Read a parameter file as txt, yaml or json

    Returns:
        dict() with parameter names as keys. Empty dictionary if
            no parameters in the file.

    Raises ValueError or IOError if no files are readable
    """
    params_dict = None

    if not Path(filename).exists():
        raise FileNotFoundError(str(filename) + " not found")

    if not Path(filename).read_text(encoding="utf-8").strip():
        logger.warning("%s was empty", filename)
        return {}

    yaml_error = ""
    try:
        logger.debug("Trying to parse %s with yaml.safe_load()", filename)
        params_dict = yaml.safe_load(Path(filename).read_text(encoding="utf-8"))
        logger.debug(" - ok, parsed as yaml")
        if not isinstance(params_dict, dict):
            # yaml happily parses txt files into a single line, don't want that.
            params_dict = None
    except Exception as yaml_exc:
        yaml_error = str(yaml_exc)
        logger.debug("%s was not parseable with yaml, trying json.", filename)

    json_error = ""
    if not params_dict:
        try:
            logger.debug("Trying to parse %s with json.load()", filename)
            with open(filename, encoding="utf-8") as f_handle:
                params_dict = json.load(f_handle)
            assert isinstance(params_dict, dict)
            logger.debug(" - ok, parsed as yaml")
        except Exception as json_exc:
            json_error = str(json_exc)
            logger.debug("%s was not parseable with json, trying txt.", filename)

    txt_error = ""
    if not params_dict:
        try:
            logger.debug("Trying to parse %s as txt with pd.read_csv()", filename)
            params_dict = load_parameterstxt(filename)
            assert isinstance(params_dict, dict)
            logger.debug(" - ok, parsed as txt")
        except Exception as txt_exc:
            txt_error = str(txt_exc)
            logger.debug("%s was not parseable as txt, no more options", filename)

    if not params_dict:
        logger.warning("%s could not be parsed as yaml, json or txt", filename)
        logger.warning("%s%s%s", str(yaml_error), str(json_error), str(txt_error))
        raise ValueError(f"Could not parse {filename}")

    # Filter to values that are NOT dict's. We can have dict as value when
    # "grouped" keys are present in the json files, both as "group:key value"
    # and in a dict called group
    params_dict = {
        key: value
        for (key, value) in params_dict.items()
        if not isinstance(value, dict)
    }
    return params_dict



# pylint: disable=c0301
"""Extract FIP region reports from Eclipse PRT file"""

import argparse
import datetime
import logging
import re
from typing import List, Optional, Union

import numpy as np
import pandas as pd

from ecl2df import EclFiles, getLogger_ecl2csv
from ecl2df.common import parse_ecl_month, write_dframe_stdout_file

logger = logging.getLogger(__name__)

REGION_REPORT_COLUMNS: List[str] = [
    "DATE",
    "FIPNAME",
    "REGION",
    "DATATYPE",
    "TO_REGION",
    "STOIIP_OIL",
    "ASSOCIATEDOIL_GAS",
    "STOIIP_TOTAL",
    "WIIP_TOTAL",
    "GIIP_GAS",
    "ASSOCIATEDGAS_OIL",
    "GIIP_TOTAL",
]


def report_block_lineparser(line: str) -> tuple:
    """
    Parses single lines within region reports, splits data into a tuple.

    Does not support many different phase configurations yet.
    """

    def float_or_nan(string: str) -> float:
        try:
            return float(string)
        except ValueError:
            return np.nan

    allowed_line_starts = [":CURRENTLY", ":OUTFLOW", ":MATERIAL", ":ORIGINALLY"]
    if not any(line.strip().upper().startswith(x) for x in allowed_line_starts):
        return tuple()

    colonsections = line.split(":")
    to_index: Optional[int]
    if "OUTFLOW TO REGION" in line:
        to_index = int(colonsections[1].split()[3])
        row_name = "OUTFLOW TO REGION"
    else:
        to_index = None
        row_name = " ".join(colonsections[1].strip().upper().split())

    # Oil section:
    liquid_oil: Optional[float] = None
    vapour_oil: Optional[float] = None
    total_oil: Optional[float] = None
    if len(colonsections[2].split()) == 3:
        (liquid_oil, vapour_oil, total_oil) = map(
            float_or_nan, colonsections[2].split()
        )
    elif len(colonsections[2].split()) == 1:
        total_oil = float_or_nan(colonsections[2])
    else:
        (liquid_oil, total_oil) = map(float_or_nan, colonsections[2].split())

    total_water = float_or_nan(colonsections[3])

    # Gas section:
    free_gas = None
    dissolved_gas = None
    total_gas = None
    if len(colonsections[4].split()) == 1:
        total_gas = float_or_nan(colonsections[4])
    elif len(colonsections[4].split()) == 2:
        (free_gas, total_gas) = map(float_or_nan, colonsections[4].split())
    else:
        (free_gas, dissolved_gas, total_gas) = map(
            float_or_nan, colonsections[4].split()
        )
    return (
        row_name,
        to_index,
        liquid_oil,
        vapour_oil,
        total_oil,
        total_water,
        free_gas,
        dissolved_gas,
        total_gas,
    )


def df(prtfile: Union[str, EclFiles], fipname: str = "FIPNUM") -> pd.DataFrame:
    """
    Parses a PRT file from Eclipse and finds FIPXXXX REGION REPORT blocks and
    organizes those numbers into a dataframe

    Each row in the dataframe represents one parsed line in the PRT file, with
    DATE and region index added.

    Args:
        prtfile: filename (PRT) or an EclFiles object
        fipname: The name of the regport regions, FIPNUM, FIPZON or whatever
            Max length of the string is 8, the first three characters must be FIP,
            and the next 3 characters must be unique for a given Eclipse deck.
    """
    if isinstance(prtfile, EclFiles):
        prtfile = prtfile.get_prtfilename()
    if not fipname.startswith("FIP"):
        raise ValueError("fipname must start with FIP")
    if len(fipname) > 8:
        raise ValueError("fipname can be at most 8 characters")

    # List of rows in final dataframe
    records = []

    # State variables while parsing line by line:
    in_report_block = False
    region_index = None
    date = None

    ecl_datematcher = re.compile(r"\s\sREPORT\s+\d+\s+(\d+)\s+(\w+)\s+(\d+)")
    opm_datematcher = re.compile(r"Starting time step.*? date = (\d+)-(\w+)-(\d+)\s*")

    # When case insensitive, this one works with both Eclipse100 and OPM:
    reportblockmatcher = re.compile(
        ".+" + fipname + r"\s+REPORT\s+REGION\s+(\d+)", re.IGNORECASE
    )

    # Flag for whether we are supposedly parsing a PRT file made by OPM flow:
    opm = False

    with open(prtfile, encoding="utf-8") as prt_fh:
        logger.info(
            "Parsing file %s for blocks starting with %s REPORT REGION",
            prtfile,
            fipname,
        )
        for line in prt_fh:
            matcheddate = re.match(ecl_datematcher, line)
            if matcheddate is None:
                matcheddate = re.match(opm_datematcher, line)
                if matcheddate is not None:
                    opm = True
            if matcheddate is not None:
                newdate = datetime.date(
                    year=int(matcheddate.group(3)),
                    month=parse_ecl_month(matcheddate.group(2).upper()),
                    day=int(matcheddate.group(1)),
                )
                if newdate != date:
                    date = newdate
                    logger.debug("Found date: %s", str(date))
                continue
            matchedreportblock = re.match(reportblockmatcher, line)
            if matchedreportblock:
                in_report_block = True
                region_index = int(matchedreportblock.group(1))
                logger.debug("  Region report for region %s", str(region_index))
                continue
            if line.startswith(" ============================"):
                in_report_block = False
                continue

            if in_report_block:
                interesting_strings = ["IN PLACE", "OUTFLOW", "MATERIAL"]
                if not sum([string in line.upper() for string in interesting_strings]):
                    # Skip if we are not on an interesting line.
                    continue

                if opm is False:
                    # The colons in the report block are not reliably included
                    # (differs by Eclipse version), even in the same PRT file. We
                    # insert them in fixed positions and hope for the best (if the
                    # ASCII table is actually dynamic with respect to content, this
                    # will fail)
                    linechars = list(line)
                    linechars[1] = ":"
                    linechars[27] = ":"
                    line = "".join(linechars)

                records.append(
                    [date, fipname, region_index] + list(report_block_lineparser(line))
                )
    return pd.DataFrame(data=records, columns=REGION_REPORT_COLUMNS)


def fill_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Fill parser with command line arguments"""
    parser.add_argument("PRTFILE", type=str, help="Eclipse PRT file (or DATA file)")
    parser.add_argument(
        "--fipname",
        type=str,
        help="Region parameter name of interest",
        default="FIPNUM",
    )
    parser.add_argument(
        "-o", "--output", type=str, help="Output CSV filename", default="outflow.csv"
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose")
    parser.add_argument("--debug", action="store_true", help="Debug mode for logging")
    return parser


def fipreports_main(args) -> None:
    """Command line API"""
    logger = getLogger_ecl2csv(  # pylint: disable=redefined-outer-name
        __name__, vars(args)
    )
    if args.PRTFILE.endswith(".PRT"):
        prtfile = args.PRTFILE
    else:
        prtfile = EclFiles(args.PRTFILE).get_prtfilename()
    dframe = df(prtfile, args.fipname)
    write_dframe_stdout_file(dframe, args.output, index=False, caller_logger=logger)



import datetime

import pandas as pd
import pytest

from ecl2df import EclFiles, compdat

try:
    import opm  # noqa
except ImportError:
    pytest.skip(
        "OPM is not installed",
        allow_module_level=True,
    )


@pytest.mark.parametrize(
    "deckstr, expected_df",
    [
        (
            """
    DATES
     1 MAY 2001 /
    /

    WLIST
     '*OP' NEW OP1 /
    /
    """,
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(2001, 5, 1),
                    }
                ]
            ),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            # Test initialization of list with no wells
            """
    DATES
     1 MAY 2001 /
    /

    WLIST
     '*OP' NEW /
    /
    """,
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "",
                        "DATE": datetime.date(2001, 5, 1),
                    }
                ]
            ),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            """
    DATES
     1 MAY 2001 /
    /

    WLIST
     '*OP' NEW OP1 /
    /

    DATES
     2 MAY 2001 /
    /
    WLIST
      '*OP' ADD OP2 OP3 /
    /
    """,
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(2001, 5, 1),
                    },
                    {
                        "NAME": "OP",
                        "ACTION": "ADD",
                        "WELLS": "OP2 OP3",
                        "DATE": datetime.date(2001, 5, 2),
                    },
                ]
            ),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
        # Construct well list from existing well list
        (
            """
    DATES
     1 MAY 2001 /
    /

    WLIST
     '*OP' NEW OP1 /
    /

    DATES
     2 MAY 2001 /
    /
    WLIST
      '*OPS' NEW /
      '*OPS' ADD '*OP' /
    /
    """,
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(2001, 5, 1),
                    },
                    {
                        "NAME": "OPS",
                        "ACTION": "NEW",
                        "WELLS": "",
                        "DATE": datetime.date(2001, 5, 2),
                    },
                    {
                        "NAME": "OPS",
                        "ACTION": "ADD",
                        "WELLS": "*OP",
                        "DATE": datetime.date(2001, 5, 2),
                    },
                ]
            ),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
    ],
)
def test_parse_wlist(deckstr, expected_df):
    deck = EclFiles.str2deck(deckstr)
    wlistdf = compdat.deck2dfs(deck)["WLIST"]
    pd.testing.assert_frame_equal(wlistdf, expected_df, check_like=True)


@pytest.mark.parametrize(
    "wlist_df, expected_df",
    [
        pytest.param(
            pd.DataFrame([{}]),
            pd.DataFrame([{}]),
            # marks=pytest.mark.xfail(raises=KeyError),
        ),
        (
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP2 OP1",
                        "DATE": datetime.date(1900, 1, 1),
                    }
                ]
            ),
            # Only change is that well names are sorted:
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP1 OP2",
                        "DATE": datetime.date(1900, 1, 1),
                    }
                ]
            ),
        ),
        (
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                    {
                        "NAME": "OP",
                        "ACTION": "ADD",
                        "WELLS": "OP2",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                ]
            ),
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP1 OP2",
                        "DATE": datetime.date(1900, 1, 1),
                    }
                ]
            ),
        ),
        (
            # Same as above, but split on two dates:
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                    {
                        "NAME": "OP",
                        "ACTION": "ADD",
                        "WELLS": "OP2",
                        "DATE": datetime.date(1900, 1, 2),
                    },
                ]
            ),
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP1 OP2",
                        "DATE": datetime.date(1900, 1, 2),
                    },
                ]
            ),
        ),
        (
            # Existing well-lists will be repeated on subsequent dates:
            pd.DataFrame(
                [
                    {
                        "NAME": "OPA",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                    {
                        "NAME": "OPB",
                        "ACTION": "NEW",
                        "WELLS": "OP2",
                        "DATE": datetime.date(1900, 1, 2),
                    },
                ]
            ),
            pd.DataFrame(
                [
                    {
                        "NAME": "OPA",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                    {
                        "NAME": "OPA",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(1900, 1, 2),
                    },
                    {
                        "NAME": "OPB",
                        "ACTION": "NEW",
                        "WELLS": "OP2",
                        "DATE": datetime.date(1900, 1, 2),
                    },
                ]
            ),
        ),
        (
            # Subsequent NEW statements with empty list clears existing
            # wells from the well list, here on the same date:
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                ]
            ),
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                ]
            ),
        ),
        (
            # Subsequent NEW statements with empty list clears existing
            # wells from the well list:
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "",
                        "DATE": datetime.date(1900, 1, 2),
                    },
                ]
            ),
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "",
                        "DATE": datetime.date(1900, 1, 2),
                    },
                ]
            ),
        ),
        (
            pd.DataFrame(
                [
                    {
                        "NAME": "OPW",
                        "ACTION": "NEW",
                        "WELLS": "OP1 OP2",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                    {
                        "NAME": "OPE",
                        "ACTION": "NEW",
                        "WELLS": "OP3 OP4",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                    {
                        "NAME": "OPC",
                        "ACTION": "MOV",
                        "WELLS": "OP2 OP3",
                        "DATE": datetime.date(1900, 1, 2),
                    },
                ]
            ),
            pd.DataFrame(
                [
                    {
                        "NAME": "OPW",
                        "ACTION": "NEW",
                        "WELLS": "OP1 OP2",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                    {
                        "NAME": "OPE",
                        "ACTION": "NEW",
                        "WELLS": "OP3 OP4",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                    {
                        "NAME": "OPW",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(1900, 1, 2),
                    },
                    {
                        "NAME": "OPE",
                        "ACTION": "NEW",
                        "WELLS": "OP4",
                        "DATE": datetime.date(1900, 1, 2),
                    },
                    {
                        "NAME": "OPC",
                        "ACTION": "NEW",
                        "WELLS": "OP2 OP3",
                        "DATE": datetime.date(1900, 1, 2),
                    },
                ]
            ),
        ),
        (
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP1 OP2 OP3",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                    {
                        "NAME": "OP",
                        "ACTION": "DEL",
                        "WELLS": "OP2",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                ]
            ),
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP1 OP3",
                        "DATE": datetime.date(1900, 1, 1),
                    }
                ]
            ),
        ),
        (
            # Adding elements from another well list, defined
            # on the same date:
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(2001, 5, 1),
                    },
                    {
                        "NAME": "OPS",
                        "ACTION": "NEW",
                        "WELLS": "",
                        "DATE": datetime.date(2001, 5, 1),
                    },
                    {
                        "NAME": "OPS",
                        "ACTION": "ADD",
                        "WELLS": "*OP",
                        "DATE": datetime.date(2001, 5, 1),
                    },
                ]
            ),
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(2001, 5, 1),
                    },
                    {
                        "NAME": "OPS",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(2001, 5, 1),
                    },
                ]
            ),
        ),
        (
            # Adding elements from another well list,
            # directly with a NEW statement:
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(2001, 5, 1),
                    },
                    {
                        "NAME": "OPS",
                        "ACTION": "NEW",
                        "WELLS": "*OP",
                        "DATE": datetime.date(2001, 5, 1),
                    },
                ]
            ),
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(2001, 5, 1),
                    },
                    {
                        "NAME": "OPS",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(2001, 5, 1),
                    },
                ]
            ),
        ),
        (
            # Adding elements from another well list,
            # directly with a NEW statement, recursively:
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(2001, 5, 1),
                    },
                    {
                        "NAME": "OPS",
                        "ACTION": "NEW",
                        "WELLS": "*OP",
                        "DATE": datetime.date(2001, 5, 1),
                    },
                    {
                        "NAME": "OPST",
                        "ACTION": "NEW",
                        "WELLS": "*OPS",
                        "DATE": datetime.date(2001, 5, 1),
                    },
                ]
            ),
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(2001, 5, 1),
                    },
                    {
                        "NAME": "OPS",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(2001, 5, 1),
                    },
                    {
                        "NAME": "OPST",
                        "ACTION": "NEW",
                        "WELLS": "OP1",
                        "DATE": datetime.date(2001, 5, 1),
                    },
                ]
            ),
        ),
        pytest.param(
            # Adding to a nonexisting list
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "ADD",
                        "WELLS": "OP1",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                ]
            ),
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        pytest.param(
            # Adding from a previously undefined list
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                    {
                        "NAME": "OP",
                        "ACTION": "ADD",
                        "WELLS": "*OPS",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                ]
            ),
            None,
            marks=pytest.mark.xfail(raises=ValueError),
            # "Recursive well list OPS does not exist"
        ),
        # Wildcard wells should pass through, the dataframe user
        # will have to process it.
        (
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "PROD*",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                ]
            ),
            pd.DataFrame(
                [
                    {
                        "NAME": "OP",
                        "ACTION": "NEW",
                        "WELLS": "PROD*",
                        "DATE": datetime.date(1900, 1, 1),
                    },
                ]
            ),
        ),
    ],
)
def test_expand_wlist(wlist_df, expected_df):
    pd.testing.assert_frame_equal(
        compdat.expand_wlist(wlist_df), expected_df, check_like=True
    )



"""Test module for fipreports"""

import datetime
import os
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from ecl2df import ecl2csv, fipreports
from ecl2df.eclfiles import EclFiles
from ecl2df.fipreports import report_block_lineparser as parser

TESTDIR = Path(__file__).absolute().parent
DATAFILE = str(TESTDIR / "data/reek/eclipse/model/2_R001_REEK-0.DATA")
MOCKPRTFILE = str(TESTDIR / "data/fipreports/TEST1.PRT")


def test_fipreports2df():
    """Test parsing of Reek dataset"""
    prtfile = EclFiles(DATAFILE).get_prtfilename()
    fipreport_df = fipreports.df(prtfile)
    assert len(fipreport_df["REGION"].unique()) == 6
    assert len(fipreport_df["DATE"].unique()) == 1
    assert fipreport_df["FIPNAME"].unique()[0] == "FIPNUM"
    assert len(fipreport_df["DATATYPE"].unique()) == 5


def test_opm_prt_file():
    """Test parsing a PRT file from OPM"""
    fipreport_df = fipreports.df(
        TESTDIR / "data/reek/eclipse/model/2_R001_REEK-0-OPMFLOW.PRT"
    )
    assert len(fipreport_df) == 530
    assert len(fipreport_df["DATE"].unique()) == 38
    assert set(fipreport_df["REGION"]) == {1, 2, 3, 4, 5, 6}
    assert set(
        [
            "DATE",
            "FIPNAME",
            "REGION",
            "DATATYPE",
            "STOIIP_OIL",
            "ASSOCIATEDOIL_GAS",
            "STOIIP_TOTAL",
            "WIIP_TOTAL",
            "GIIP_GAS",
            "ASSOCIATEDGAS_OIL",
            "GIIP_TOTAL",
        ]
    ).issubset(set(fipreport_df.columns))


def test_mockprtfile():
    """
    Test (a) mocked PRT file(s)
    """
    dframe = fipreports.df(MOCKPRTFILE)
    assert dframe["FIPNAME"].unique() == "FIPNUM"
    assert len(dframe["DATE"].unique()) == 1
    assert int(dframe.loc[0, "GIIP_GAS"]) == 20

    dframe = fipreports.df(MOCKPRTFILE, fipname="FIPZON")
    assert dframe["FIPNAME"].unique() == "FIPZON"
    assert len(dframe["REGION"].unique()) == 2
    assert len(dframe["TO_REGION"].dropna().unique()) == 4
    assert len(dframe["DATE"].unique()) == 3
    assert len(dframe["DATATYPE"].unique()) == 6

    dframe = fipreports.df(MOCKPRTFILE, fipname="FIPOWG")
    assert dframe["FIPNAME"].unique() == "FIPOWG"
    assert len(dframe["DATE"].unique()) == 1

    with pytest.raises(ValueError):
        fipreports.df(MOCKPRTFILE, fipname="WIPNUM")

    # fipname at most 8 characters:
    assert fipreports.df(MOCKPRTFILE, fipname="FIP45678").empty

    with pytest.raises(ValueError):
        fipreports.df(MOCKPRTFILE, fipname="FIP456789")


def test_prtstring(tmp_path):
    """Test a PRT from string, verifying every detail of the dataframe"""
    prtstring = """
  REPORT   0     1 JAN 2000
                                                =================================
                                                : FIPNUM  REPORT REGION    2    :
                                                :     PAV =        139.76  BARSA:
                                                :     PORV=     27777509.   RM3 :
                           :--------------- OIL    SM3  ---------------:-- WAT    SM3  -:--------------- GAS    SM3  ---------------:
                           :     LIQUID         VAPOUR         TOTAL   :       TOTAL    :       FREE      DISSOLVED         TOTAL   :
 :-------------------------:-------------------------------------------:----------------:-------------------------------------------:
 :CURRENTLY IN PLACE       :     21091398.                    21091398.:       4590182. :           -0.    483594842.     483594842.:
 :-------------------------:-------------------------------------------:----------------:-------------------------------------------:
 :OUTFLOW TO OTHER REGIONS :        76266.                       76266.:         75906. :            0.      1818879.       1818879.:
 :OUTFLOW THROUGH WELLS    :                                         0.:             0. :                                         0.:
 :MATERIAL BALANCE ERROR.  :                                         0.:             0. :                                         0.:
 :-------------------------:-------------------------------------------:----------------:-------------------------------------------:
 :ORIGINALLY IN PLACE      :     21136892.                    21136892.:       4641214. :            0.    484657561.     484657561.:
 :-------------------------:-------------------------------------------:----------------:-------------------------------------------:
 :OUTFLOW TO REGION   1    :       143128.                      143128.:       -161400. :            0.      3017075.       3017075.:
 :OUTFLOW TO REGION   3    :       -66862.                      -66862.:        198900. :           -0.     -1198195.      -1198195.:
 :OUTFLOW TO REGION   8    :            0.                           0.:         38405. :            0.            0.             0.:
 ====================================================================================================================================
"""  # noqa
    os.chdir(tmp_path)
    Path("FOO.PRT").write_text(prtstring)
    dframe = fipreports.df("FOO.PRT")
    expected_dframe = pd.DataFrame(
        [
            {
                "DATE": datetime.date(2000, 1, 1),
                "FIPNAME": "FIPNUM",
                "REGION": 2,
                "DATATYPE": "CURRENTLY IN PLACE",
                "TO_REGION": np.nan,
                "STOIIP_OIL": 21091398.0,
                "ASSOCIATEDOIL_GAS": None,
                "STOIIP_TOTAL": 21091398.0,
                "WIIP_TOTAL": 4590182.0,
                "GIIP_GAS": -0.0,
                "ASSOCIATEDGAS_OIL": 483594842.0,
                "GIIP_TOTAL": 483594842.0,
            },
            {
                "DATE": datetime.date(2000, 1, 1),
                "FIPNAME": "FIPNUM",
                "REGION": 2,
                "DATATYPE": "OUTFLOW TO OTHER REGIONS",
                "TO_REGION": np.nan,
                "STOIIP_OIL": 76266.0,
                "ASSOCIATEDOIL_GAS": None,
                "STOIIP_TOTAL": 76266.0,
                "WIIP_TOTAL": 75906.0,
                "GIIP_GAS": 0.0,
                "ASSOCIATEDGAS_OIL": 1818879.0,
                "GIIP_TOTAL": 1818879.0,
            },
            {
                "DATE": datetime.date(2000, 1, 1),
                "FIPNAME": "FIPNUM",
                "REGION": 2,
                "DATATYPE": "OUTFLOW THROUGH WELLS",
                "TO_REGION": np.nan,
                "STOIIP_OIL": np.nan,
                "ASSOCIATEDOIL_GAS": None,
                "STOIIP_TOTAL": 0.0,
                "WIIP_TOTAL": 0.0,
                "GIIP_GAS": np.nan,
                "ASSOCIATEDGAS_OIL": np.nan,
                "GIIP_TOTAL": 0.0,
            },
            {
                "DATE": datetime.date(2000, 1, 1),
                "FIPNAME": "FIPNUM",
                "REGION": 2,
                "DATATYPE": "MATERIAL BALANCE ERROR.",
                "TO_REGION": np.nan,
                "STOIIP_OIL": np.nan,
                "ASSOCIATEDOIL_GAS": None,
                "STOIIP_TOTAL": 0.0,
                "WIIP_TOTAL": 0.0,
                "GIIP_GAS": np.nan,
                "ASSOCIATEDGAS_OIL": np.nan,
                "GIIP_TOTAL": 0.0,
            },
            {
                "DATE": datetime.date(2000, 1, 1),
                "FIPNAME": "FIPNUM",
                "REGION": 2,
                "DATATYPE": "ORIGINALLY IN PLACE",
                "TO_REGION": np.nan,
                "STOIIP_OIL": 21136892.0,
                "ASSOCIATEDOIL_GAS": None,
                "STOIIP_TOTAL": 21136892.0,
                "WIIP_TOTAL": 4641214.0,
                "GIIP_GAS": 0.0,
                "ASSOCIATEDGAS_OIL": 484657561.0,
                "GIIP_TOTAL": 484657561.0,
            },
            {
                "DATE": datetime.date(2000, 1, 1),
                "FIPNAME": "FIPNUM",
                "REGION": 2,
                "DATATYPE": "OUTFLOW TO REGION",
                "TO_REGION": 1.0,
                "STOIIP_OIL": 143128.0,
                "ASSOCIATEDOIL_GAS": None,
                "STOIIP_TOTAL": 143128.0,
                "WIIP_TOTAL": -161400.0,
                "GIIP_GAS": 0.0,
                "ASSOCIATEDGAS_OIL": 3017075.0,
                "GIIP_TOTAL": 3017075.0,
            },
            {
                "DATE": datetime.date(2000, 1, 1),
                "FIPNAME": "FIPNUM",
                "REGION": 2,
                "DATATYPE": "OUTFLOW TO REGION",
                "TO_REGION": 3.0,
                "STOIIP_OIL": -66862.0,
                "ASSOCIATEDOIL_GAS": None,
                "STOIIP_TOTAL": -66862.0,
                "WIIP_TOTAL": 198900.0,
                "GIIP_GAS": -0.0,
                "ASSOCIATEDGAS_OIL": -1198195.0,
                "GIIP_TOTAL": -1198195.0,
            },
            {
                "DATE": datetime.date(2000, 1, 1),
                "FIPNAME": "FIPNUM",
                "REGION": 2,
                "DATATYPE": "OUTFLOW TO REGION",
                "TO_REGION": 8.0,
                "STOIIP_OIL": 0.0,
                "ASSOCIATEDOIL_GAS": None,
                "STOIIP_TOTAL": 0.0,
                "WIIP_TOTAL": 38405.0,
                "GIIP_GAS": 0.0,
                "ASSOCIATEDGAS_OIL": 0.0,
                "GIIP_TOTAL": 0.0,
            },
        ]
    )

    pd.testing.assert_frame_equal(dframe, expected_dframe)


def test_drygas_report(tmp_path):
    """Excerpt from a two-phase gas water run"""
    prtstring = """
                                                =================================
                                                : FIPNUM  REPORT REGION    2    :
                                                :     PAV =        909.34  BARSA:
                                                :     PORV=    150001895.   RM3 :
                           :--------------- OIL    SM3  ---------------:-- WAT    SM3  -:--------------- GAS    SM3  ---------------:
                           :     LIQUID         VAPOUR         TOTAL   :       TOTAL    :       FREE      DISSOLVED         TOTAL   :
 :-------------------------:-------------------------------------------:----------------:-------------------------------------------:
 :CURRENTLY IN PLACE       :            0.                           0.:      10476036. :   1815774165.                  1815774165.:
 :-------------------------:-------------------------------------------:----------------:-------------------------------------------:
 :OUTFLOW TO OTHER REGIONS :            0.                           0.:             0. :            0.                           0.:
 :OUTFLOW THROUGH WELLS    :                                         0.:             0. :                                         0.:
 :MATERIAL BALANCE ERROR.  :                                        -0.:             0. :                                         0.:
 :-------------------------:-------------------------------------------:----------------:-------------------------------------------:
 :ORIGINALLY IN PLACE      :            0.                           0.:      10476036. :   1815774165.                  1815774165.:
 :-------------------------:-------------------------------------------:----------------:-------------------------------------------:
 ====================================================================================================================================
"""  # noqa
    os.chdir(tmp_path)
    Path("FOO.PRT").write_text(prtstring)
    dframe = fipreports.df("FOO.PRT").set_index("DATATYPE")
    assert dframe["REGION"].unique() == [2]
    pd.testing.assert_frame_equal(
        dframe.reset_index().drop(["DATE", "REGION"], axis="columns"),
        pd.DataFrame(
            [
                {
                    "DATATYPE": "CURRENTLY IN PLACE",
                    "FIPNAME": "FIPNUM",
                    "TO_REGION": None,
                    "STOIIP_OIL": 0.0,
                    "ASSOCIATEDOIL_GAS": None,
                    "STOIIP_TOTAL": 0.0,
                    "WIIP_TOTAL": 10476036.0,
                    "GIIP_GAS": 1815774165.0,
                    "ASSOCIATEDGAS_OIL": None,
                    "GIIP_TOTAL": 1815774165.0,
                },
                {
                    "DATATYPE": "OUTFLOW TO OTHER REGIONS",
                    "FIPNAME": "FIPNUM",
                    "TO_REGION": None,
                    "STOIIP_OIL": 0.0,
                    "ASSOCIATEDOIL_GAS": None,
                    "STOIIP_TOTAL": 0.0,
                    "WIIP_TOTAL": 0.0,
                    "GIIP_GAS": 0.0,
                    "ASSOCIATEDGAS_OIL": None,
                    "GIIP_TOTAL": 0.0,
                },
                {
                    "DATATYPE": "OUTFLOW THROUGH WELLS",
                    "FIPNAME": "FIPNUM",
                    "TO_REGION": None,
                    "STOIIP_OIL": np.nan,
                    "ASSOCIATEDOIL_GAS": None,
                    "STOIIP_TOTAL": 0.0,
                    "WIIP_TOTAL": 0.0,
                    "GIIP_GAS": np.nan,
                    "ASSOCIATEDGAS_OIL": None,
                    "GIIP_TOTAL": 0.0,
                },
                {
                    "DATATYPE": "MATERIAL BALANCE ERROR.",
                    "FIPNAME": "FIPNUM",
                    "TO_REGION": None,
                    "STOIIP_OIL": np.nan,
                    "ASSOCIATEDOIL_GAS": None,
                    "STOIIP_TOTAL": -0.0,
                    "WIIP_TOTAL": 0.0,
                    "GIIP_GAS": np.nan,
                    "ASSOCIATEDGAS_OIL": None,
                    "GIIP_TOTAL": 0.0,
                },
                {
                    "DATATYPE": "ORIGINALLY IN PLACE",
                    "FIPNAME": "FIPNUM",
                    "TO_REGION": None,
                    "STOIIP_OIL": 0.0,
                    "ASSOCIATEDOIL_GAS": None,
                    "STOIIP_TOTAL": 0.0,
                    "WIIP_TOTAL": 10476036.0,
                    "GIIP_GAS": 1815774165.0,
                    "ASSOCIATEDGAS_OIL": None,
                    "GIIP_TOTAL": 1815774165.0,
                },
            ]
        ),
    )


def test_rogue_eclipse_output(tmp_path):
    """The stars in the material balance error line has been observed in reality."""
    prtstring = """
                                                =================================
                                                : FIPNUM  REPORT REGION  120    :
                                                :     PAV =        298.89  BARSA:
                                                :     PORV=      4502843.   RM3 :
                           :--------------- OIL    SM3  ---------------:-- WAT    SM3  -:--------------- GAS    SM3  ---------------:
                           :     LIQUID         VAPOUR         TOTAL   :       TOTAL    :       FREE      DISSOLVED         TOTAL   :
 :-------------------------:-------------------------------------------:----------------:-------------------------------------------:
 :CURRENTLY IN PLACE       :     -2703242.        10451.      -2692791.:       2568336. :     59233087. 190842667352.  190901900439.:
 :-------------------------:-------------------------------------------:----------------:-------------------------------------------:
 :OUTFLOW TO OTHER REGIONS :       294586.         6362.        300947.:       1235671. :     39452538.     51855907.      91308445.:
 :OUTFLOW THROUGH WELLS    :                                     65430.:      -1818966. :                                 -85526625.:
 :MATERIAL BALANCE ERROR.  :                                   3419391.:        671761. :                              *************:
 :-------------------------:-------------------------------------------:----------------:-------------------------------------------:
"""  # noqa
    os.chdir(tmp_path)
    Path("FOO.PRT").write_text(prtstring)
    dframe = fipreports.df("FOO.PRT").set_index("DATATYPE")
    assert np.isnan(dframe.loc["MATERIAL BALANCE ERROR.", "GIIP_TOTAL"])


def test_prtstring_opmflow(tmp_path):
    """Test parsing the PRT output from OPM flow."""
    prtstring = """
Starting time step 3, stepsize 19.6 days, at day 11.4/31, date = 12-Jan-2000

                                                  ===================================================
                                                  :        FIPNUM report region   1                 :
                                                  :      PAV  =       306.192 BARSA                 :
                                                  :      PORV =      78804306   RM3                 :
                         :--------------- Oil    SM3 ---------------:-- Wat    SM3 --:--------------- Gas    SM3 ---------------:
                         :      Liquid        Vapour        Total   :      Total     :      Free        Dissolved       Total   :
:------------------------:------------------------------------------:----------------:------------------------------------------:
:Currently   in place    :      16528782             0      16528782:     60416351   :             0             0             0:
:------------------------:------------------------------------------:----------------:------------------------------------------:
:Originally  in place    :      16530271             0      16530271:     60415965   :             0             0             0:
:========================:==========================================:================:==========================================:
"""  # noqa
    os.chdir(tmp_path)
    Path("FOO.PRT").write_text(prtstring)
    dframe = fipreports.df("FOO.PRT")
    print(dframe.to_dict(orient="records"))
    expected_dframe = pd.DataFrame(
        [
            {
                "DATE": datetime.date(2000, 1, 12),
                "FIPNAME": "FIPNUM",
                "REGION": 1,
                "DATATYPE": "CURRENTLY IN PLACE",
                "TO_REGION": None,
                "STOIIP_OIL": 16528782.0,
                "ASSOCIATEDOIL_GAS": 0.0,
                "STOIIP_TOTAL": 16528782.0,
                "WIIP_TOTAL": 60416351.0,
                "GIIP_GAS": 0.0,
                "ASSOCIATEDGAS_OIL": 0.0,
                "GIIP_TOTAL": 0.0,
            },
            {
                "DATE": datetime.date(2000, 1, 12),
                "FIPNAME": "FIPNUM",
                "REGION": 1,
                "DATATYPE": "ORIGINALLY IN PLACE",
                "TO_REGION": None,
                "STOIIP_OIL": 16530271.0,
                "ASSOCIATEDOIL_GAS": 0.0,
                "STOIIP_TOTAL": 16530271.0,
                "WIIP_TOTAL": 60415965.0,
                "GIIP_GAS": 0.0,
                "ASSOCIATEDGAS_OIL": 0.0,
                "GIIP_TOTAL": 0.0,
            },
        ]
    )
    pd.testing.assert_frame_equal(dframe, expected_dframe)


def test_report_block_lineparser():
    """
    Test the line-parser, which has to infer partly which phases are present.
    """

    tup = parser(
        " :OUTFLOW THROUGH WELLS    :                                         0.:"
        "             0. :                                         0.:"
    )
    assert tup[0] == "OUTFLOW THROUGH WELLS"
    assert not tup[1]  # to_region is N/A
    assert not tup[2]
    assert not tup[3]
    assert tup[4] == 0.0
    assert tup[5] == 0.0
    assert not tup[6]
    assert not tup[7]
    assert tup[8] == 0.0

    assert not parser("")
    assert not parser("foobarcom")

    tup = parser(
        " :CURRENTLY IN PLACE       :    610520867.                   610520867.:"
        "    5096189976. :            0.  22298026321.   22298026321.:"
    )
    assert tup[0] == "CURRENTLY IN PLACE"
    assert not tup[1]  # to_region is N/A
    assert int(tup[2]) == 610520867
    assert not tup[3]
    assert int(tup[6]) == 0
    assert int(tup[7]) == 22298026321


def test_cmdline(tmp_path, mocker):
    """Test command line interface"""
    tmpcsvfile = tmp_path / "TMP-fipreports.csv"
    mocker.patch(
        "sys.argv",
        ["ecl2csv", "fipreports", "-v", DATAFILE, "--output", str(tmpcsvfile)],
    )
    ecl2csv.main()

    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(tmpcsvfile)
    assert "FIPNAME" in disk_df
    assert "STOIIP_OIL" in disk_df
    assert not disk_df.empty

    # Debug mode:
    mocker.patch(
        "sys.argv",
        [
            "ecl2csv",
            "fipreports",
            "--debug",
            DATAFILE,
            "--output",
            "debugmode.csv",
        ],
    )
    ecl2csv.main()
    pd.testing.assert_frame_equal(pd.read_csv("debugmode.csv"), disk_df)

    # Directly on PRT file:
    mocker.patch(
        "sys.argv",
        [
            "ecl2csv",
            "fipreports",
            DATAFILE.replace("DATA", "PRT"),
            "--output",
            "fromprtfile.csv",
        ],
    )
    ecl2csv.main()
    pd.testing.assert_frame_equal(pd.read_csv("fromprtfile.csv"), disk_df)



#!/usr/bin/env python
"""
Extract grid information from Eclipse output files as Dataframes.

Each cell in the grid correspond to one row.

For grid cells, x, y, z for cell centre and volume is available as
geometric information. Static data (properties) can be merged from
the INIT file, and dynamic data can be merged from the Restart (UNRST)
file.
"""
import argparse
import datetime
import fnmatch
import logging
import textwrap
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Type, Union

import dateutil.parser
import numpy as np
import pandas as pd
import pyarrow
import pyarrow.feather
from ecl.eclfile import EclFile

from ecl2df import __version__, common, getLogger_ecl2csv

from .eclfiles import EclFiles

logger = logging.getLogger(__name__)


def get_available_rst_dates(eclfiles: EclFiles) -> List[datetime.date]:
    """Return a list of datetime objects for the available dates in the RST file"""
    report_indices = EclFile.file_report_list(eclfiles.get_rstfilename())
    logger.info(
        "Restart report indices (count %s): %s",
        str(len(report_indices)),
        str(report_indices),
    )
    return [
        eclfiles.get_rstfile().iget_restart_sim_time(index).date()
        for index in range(0, len(report_indices))
    ]


def dates2rstindices(
    eclfiles: EclFiles, dates: Optional[Union[str, datetime.date, List[datetime.date]]]
) -> Tuple[List[int], List[datetime.date], List[str]]:
    """Return the restart index/indices for a given datetime or list of datetimes

      dates: datetime.date or list of datetime.date, must
          correspond to an existing date. If list, it
          forces dateinheaders to be True.
          Can also be string, then the mnenomics
          'first', 'last', 'all', are supported, or ISO
          date formats. If None or empty string is supplied,
          an empty return tuple is returned.

    Raises exception if no dates are not found.

    Return: tuple, first element is
        list of integers, corresponding to restart indices. Length 1 or more.
        second element is list of corresponding datetime.date objecs, third
        is an ISO-8601 representation of the dates.
    """
    if not dates:
        return ([], [], [])

    availabledates = get_available_rst_dates(eclfiles)

    supportedmnemonics = ["first", "last", "all"]

    # After this control block, chosendates is a list of dates
    # we should extract, and which exists in UNRST.
    if isinstance(dates, str):
        if dates not in supportedmnemonics:
            # Try to parse as ISO date:
            try:
                isodate = dateutil.parser.isoparse(dates).date()
            except ValueError as dateparser_error:
                raise ValueError(
                    "date " + str(dates) + " not understood"
                ) from dateparser_error
            if isodate not in availabledates:
                raise ValueError("date " + str(isodate) + " not found in UNRST file")
            chosendates = [isodate]
        else:
            if dates == "first":
                chosendates = [availabledates[0]]
            elif dates == "last":
                chosendates = [availabledates[-1]]
            elif dates == "all":
                chosendates = availabledates
    elif isinstance(dates, datetime.datetime):
        chosendates = [dates.date()]
    elif isinstance(dates, datetime.date):
        # Beware, datetime.datetime would also match here.
        chosendates = [dates]
    elif isinstance(dates, list):
        chosendates = [x for x in dates if x in availabledates]
        if not chosendates:
            raise ValueError("None of the requested dates were found")
        if len(chosendates) < len(availabledates):
            logger.warning("Not all dates found in UNRST\n")
    else:
        raise ValueError("date " + str(dates) + " not understood")

    logger.info(
        "Available dates (count %s) in RST: %s",
        str(len(availabledates)),
        str([x.isoformat() for x in availabledates]),
    )
    rstindices = [availabledates.index(x) for x in chosendates]
    isostrings = [x.isoformat() for x in chosendates]
    return (rstindices, chosendates, isostrings)


def _df2pyarrow(dframe: pd.DataFrame) -> pyarrow.Table:
    """Construct a pyarrow table from dataframe with
    grid information

    The index in the dataframe will be ignored

    32-bit types will be used for integers and floats (todo)
    """
    field_list: List[pyarrow.Field] = []
    for colname in dframe.columns:
        if pd.api.types.is_integer_dtype(dframe.dtypes[colname]):
            dtype = pyarrow.int32()
        elif pd.api.types.is_string_dtype(dframe.dtypes[colname]):
            # Parameters are potentially merged into the dataframe.
            dtype = pyarrow.string()
        else:
            dtype = pyarrow.float32()
        field_list.append(pyarrow.field(colname, dtype))

    schema = pyarrow.schema(field_list)
    return pyarrow.Table.from_pandas(dframe, schema=schema, preserve_index=False)


def rst2df(
    eclfiles: EclFiles,
    date: Union[str, datetime.date, List[datetime.date]],
    vectors: Optional[Union[str, List[str]]] = None,
    dateinheaders: bool = False,
    stackdates: bool = False,
) -> pd.DataFrame:
    """Return a dataframe with dynamic data from the restart file
    for each cell, at a particular date.

    The dataframe will have a dummy index. The column named
    "active" refers to the active cell index, and is to be used
    when merging with the grid geometry dataframe.

    Args:
        eclfiles: EclFiles object
        date: datetime.date or list of datetime.date, must
            correspond to an existing date. If list, it
            forces dateinheaders to be True.
            Can also be string, then the mnenomics
            'first', 'last', 'all', are supported, or ISO
            date formats.
         vectors: List of vectors to include,
            glob-style wildcards supported
         dateinheaders: boolean on whether the date should
            be added to the column headers. Instead of
            SGAS as a column header, you get SGAS@YYYY-MM-DD.
         stackdates: Default is false. If True, a column
            called DATE will be added and data for all restart
            dates will be added in a stacked manner. Implies
            dateinheaders False.
    """
    if not vectors:
        vectorswasdefaulted = True
        vectors = "*"  # This will include everything
    else:
        vectorswasdefaulted = False

    if not isinstance(vectors, list):
        vectors = [vectors]

    # First task is to determine the restart index to extract
    # data for:
    (rstindices, chosendates, isodates) = dates2rstindices(eclfiles, date)

    logger.info("Extracting restart information at dates %s", str(isodates))

    # Determine the available restart vectors, we only include
    # those with correct length, meaning that they are defined
    # for all active cells:
    activecells = eclfiles.get_egrid().getNumActive()
    rstvectors = []
    for vec in eclfiles.get_rstfile().headers:
        if vec[1] == activecells and any(
            fnmatch.fnmatch(vec[0], key) for key in vectors
        ):
            rstvectors.append(vec[0])
    rstvectors = list(set(rstvectors))  # Make unique list
    # Note that all of these might not exist at all timesteps.

    if stackdates and dateinheaders:
        logger.warning("Will not put date in headers when stackdates=True")
        dateinheaders = False

    rst_dfs = {}
    for rstindex in rstindices:
        # Filter the rst vectors once more, all of them
        # might not be available at all timesteps:
        present_rstvectors = []
        for vec in rstvectors:
            try:
                if eclfiles.get_rstfile().iget_named_kw(vec, rstindex):
                    present_rstvectors.append(vec)
            except IndexError:
                pass
        logger.info(
            "Present restart vectors at index %s: %s",
            str(rstindex),
            str(present_rstvectors),
        )
        if not present_rstvectors:
            if vectorswasdefaulted:
                logger.warning(
                    "No restart vectors available at index %s", str(rstindex)
                )
            continue

        # Make the dataframe
        rst_df = pd.DataFrame(
            columns=present_rstvectors,
            data=np.hstack(
                [
                    eclfiles.get_rstfile()
                    .iget_named_kw(vec, rstindex)
                    .numpyView()
                    .reshape(-1, 1)
                    for vec in present_rstvectors
                ]
            ),
        )

        # For users convenience:
        if (
            "SWAT" in rst_df
            and "SGAS" in rst_df
            and "SOIL" not in rst_df
            and any(fnmatch.fnmatch("SOIL", key) for key in vectors)
        ):
            rst_df["SOIL"] = 1 - rst_df["SWAT"] - rst_df["SGAS"]

        # Tag the column names if requested, or if multiple rst indices
        # are asked for
        datestr = chosendates[rstindices.index(rstindex)].isoformat()
        if dateinheaders or len(rstindices) > 1 and not stackdates:
            rst_df.columns = [colname + "@" + datestr for colname in rst_df.columns]

        # libecl emits a number around -1.0000000200408773e+20 which
        # should be considered Not-a-number
        rst_df = rst_df.where(rst_df > -1e20 + 1e13)  # some trial and error

        # Remove columns that are all NaN:
        rst_df.dropna(axis="columns", how="all", inplace=True)

        rst_df.index.name = "active"

        rst_dfs[datestr] = rst_df

    if not rst_dfs:
        return pd.DataFrame()

    if not stackdates:
        return pd.concat(rst_dfs.values(), axis=1).reset_index()

    rststack = pd.concat(rst_dfs, sort=False).reset_index()
    rststack.rename(columns={"level_0": "DATE"}, inplace=True)
    return rststack


def gridgeometry2df(
    eclfiles: EclFiles, zonemap: Optional[Dict[int, str]] = None
) -> pd.DataFrame:
    """Produce a Pandas Dataframe with Eclipse gridgeometry

    Order is significant, and is determined by the order from libecl, and used
    when merging with other dataframes with cell-data.

    Args:
        eclfiles: object holding the Eclipse output files.
        zonemap: A zonemap dictionary mapping every K index to a
            string, which will be put in a column ZONE. If none is provided,
            a zonemap from a default file will be looked for. Provide an empty
            dictionary to avoid looking for the default file, and no ZONE
            column will be added.

    Returns:
        DataFrame with at least the columns I, J, K, X, Y, Z and VOLUME. One row
        pr. cell. The index of the dataframe are the global indices. If a zonemap
        is provided, zone information will be in the column ZONE.
    """
    egrid_file = eclfiles.get_egridfile()
    grid = eclfiles.get_egrid()

    if not egrid_file or not grid:
        raise ValueError("No EGRID file supplied")

    logger.info("Extracting grid geometry from %s", str(egrid_file))
    index_frame = grid.export_index(active_only=True)
    ijk = index_frame.values[:, 0:3] + 1  # ijk from ecl.grid is off by one

    xyz = grid.export_position(index_frame)
    vol = grid.export_volume(index_frame)
    z_corners = grid.export_corners(index_frame)[:, [2, 5, 8, 11, 14, 17, 20]]
    grid_df = pd.DataFrame(
        index=index_frame["active"],
        columns=[
            "i",
            "j",
            "k",
            "x",
            "y",
            "z",
            "z_min",
            "z_max",
            "volume",
            "global_index",
        ],
        data=np.hstack(
            (
                ijk,
                xyz,
                z_corners.min(axis=1).reshape(-1, 1),
                z_corners.max(axis=1).reshape(-1, 1),
                vol.reshape(-1, 1),
                np.array(index_frame.index).reshape(-1, 1),
            )
        ),
    )

    # Type conversion, hstack maybe ruined the datatypes..
    grid_df["i"] = grid_df["i"].astype(int)
    grid_df["j"] = grid_df["j"].astype(int)
    grid_df["k"] = grid_df["k"].astype(int)

    # Column names should be uppercase
    grid_df.columns = [x.upper() for x in grid_df.columns]

    if zonemap is None:
        # Look for default zonemap file:
        zonemap = eclfiles.get_zonemap()
    if zonemap:
        logger.info("Merging zonemap into grid")
        grid_df = common.merge_zones(grid_df, zonemap, kname="K")

    return grid_df


def merge_initvectors(
    eclfiles: EclFiles,
    dframe: pd.DataFrame,
    initvectors: List[str],
    ijknames: List[str] = None,
) -> pd.DataFrame:
    """Merge in INIT vectors to a dataframe by I, J, K.

    Utility function for other modules to use. Normally it is sufficient
    for API users to only use the df() function.

    Args:
        eclfiles: Object representing the Eclipse output files
        dframe: Table data to merge with
        initvectors: Names of INIT vectors to merge in.
        ijknames: Three strings that determine the I, J and K columns to use
            for merging in dframe. For compdat the 'K' is f.ex. denoted 'K1'

    Returns:
        Copy of incoming dataframe with additional columns
    """
    if not initvectors:
        # Nothing to do.
        return dframe
    if not ijknames:
        ijknames = ["I", "J", "K"]
    if len(ijknames) != 3:
        raise ValueError("ijknames must be a list of length 3")
    assert isinstance(dframe, pd.DataFrame)
    assert isinstance(eclfiles, EclFiles)

    if not set(ijknames).issubset(dframe.columns):
        raise ValueError(
            f"All of the columns {ijknames} must be present "
            "in the dataframe for merging"
        )

    if isinstance(initvectors, str):
        initvectors = [initvectors]
    assert isinstance(initvectors, list)

    logger.info("Merging INIT data %s into dataframe", str(initvectors))
    ijkinit = df(eclfiles, vectors=initvectors)[["I", "J", "K"] + initvectors]
    return pd.merge(dframe, ijkinit, left_on=ijknames, right_on=["I", "J", "K"])


def init2df(eclfiles: EclFiles, vectors: Union[str, List[str]] = None) -> pd.DataFrame:
    """Extract information from INIT file with cell data

    Normally, you would use the df() function which will
    call this function on your behalf.

    Order is significant, as index is used for merging

    Args:
        eclfiles: Object that can serve the EGRID and INIT files
        vectors: List of vectors to include,
            glob-style wildcards supported.
    """
    if not vectors:
        vectors = "*"  # This will include everything
    if not isinstance(vectors, list):
        vectors = [vectors]

    init = eclfiles.get_initfile()
    egrid = eclfiles.get_egrid()

    # Build list of vector names to include:
    usevectors = []
    include_porv = False
    for vec in init.headers:
        if vec[1] == egrid.getNumActive() and any(
            fnmatch.fnmatch(vec[0], key) for key in vectors
        ):
            usevectors.append(vec[0])
        if vec[0] == "PORV" and any(fnmatch.fnmatch("PORV", key) for key in vectors):
            include_porv = True

    if usevectors:
        init_df = pd.DataFrame(
            columns=usevectors,
            data=np.hstack(
                [
                    init.iget_named_kw(vec, 0).numpyView().reshape(-1, 1)
                    for vec in usevectors
                ]
            ),
        )
        # libecl emits a number around -1.0000000200408773e+20 which
        # should be considered Not-a-number
        init_df = init_df.where(init_df > -1e20 + 1e13)  # some trial and error

        # Remove columns that are all NaN:
        init_df.dropna(axis="columns", how="all", inplace=True)

    else:
        init_df = pd.DataFrame()  # empty

    logger.info("Extracted %s from INIT file", str(init_df.columns.values))

    # PORV is indexed by active_index, not global, needs special treatment:
    if include_porv:
        porv_numpy = init.iget_named_kw("PORV", 0).numpyView()
        glob_idxs = [
            egrid.get_global_index(active_index=ix)
            for ix in range(egrid.getNumActive())
        ]
        init_df["PORV"] = porv_numpy[glob_idxs].reshape(-1, 1)
    return init_df


def df(
    eclfiles: EclFiles,
    vectors: Union[str, List[str]] = "*",
    dropconstants: bool = False,
    rstdates: Optional[Union[str, datetime.date, List[datetime.date]]] = None,
    dateinheaders: bool = False,
    stackdates: bool = False,
    zonemap: Optional[Dict[int, str]] = None,
):
    """Produce a dataframe with grid information

    Grid information (center coordinates x, y, z), cell
    indices (i, j, k) (indices follow the Eclipse convention starting
    at 1, not zero as in libecl), properties from INIT, and optionally
    any time dependent data from Restart files.

    Args:
        eclfiles: Handle to an Eclipse case
        vectors: Vectors to include, wildcards
            supported. Used to match both
            INIT vectors and RESTART vectors.
        dropconstants (bool): If true, columns that are constant
            for every cell are dropped.
        rstdates: Restart dates to include in ISO-8601 format.
            Alternatively, pick from the mnenomics 'first', 'all' and 'last'.
        dateinheaders: Whether columns with data from UNRST files
            should always have the ISO-date embedded in the column header.
        stackdates: Default is false. If true, a column
            called DATE will be added and data for all restart
            dates will be added in a stacked manner. Implies
            dateinheaders False.
        zonemap: A zonemap dictionary mapping every K index to a
            string, which will be put in a column ZONE. If none is provided,
            a zonemap from a default file will be looked for. Provide an empty
            dictionary to avoid looking for the default file, and no ZONE
            column will be added.
    """
    gridgeom = gridgeometry2df(eclfiles, zonemap)
    initdf = init2df(eclfiles, vectors=vectors)
    rst_df = None
    if rstdates:
        rst_df = rst2df(
            eclfiles,
            rstdates,
            vectors=vectors,
            dateinheaders=dateinheaders,
            stackdates=stackdates,
        )
    grid_df = gridgeom.merge(
        initdf, how="outer", on=None, left_index=True, right_index=True
    )

    if rst_df is not None and not rst_df.empty:
        grid_df = grid_df.merge(
            rst_df, how="outer", left_index=True, right_on="active"
        ).reset_index(drop=True)

    if dropconstants:
        grid_df = drop_constant_columns(grid_df)
    return grid_df.drop("active", axis="columns", errors="ignore")


def fill_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Set up sys.argv parser.

    Arguments:
        parser: argparse.ArgumentParser or argparse.subparser
    """
    parser.add_argument(
        "DATAFILE",
        help="Name of Eclipse DATA file. " + "INIT and EGRID file must lie alongside.",
    )
    parser.add_argument(
        "--vectors",
        nargs="+",
        help="INIT and/or restart wildcards for vectors to include",
        default="*",
    )
    parser.add_argument(
        "--rstdates",
        type=str,
        help="Point in time to grab restart data from, "
        + "either 'first' or 'last', 'all', or a date in "
        + "YYYY-MM-DD format",
        default="",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="Name of output csv file. Use '-' for stdout.",
        default="eclgrid.csv",
    )
    parser.add_argument(
        "--stackdates",
        action="store_true",
        help=(
            "If set, the dates from restart data will not be in the column "
            "but instead there will be a DATE column with the dates. Note "
            "that the static data will be repeated for each DATE."
        ),
    )

    parser.add_argument(
        "--dropconstants",
        action="store_true",
        help="Drop constant columns from the dataset",
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose")
    parser.add_argument("--arrow", action="store_true", help="Write to pyarrow format")
    return parser


def drop_constant_columns(
    dframe: pd.DataFrame, alwayskeep: Optional[Union[str, List[str]]] = None
) -> pd.DataFrame:
    """Drop/delete constant columns from a dataframe.

    Args:
        dframe
        alwayskeep: string or list of strings of columns to keep
           anyway.
    Returns:
        Dataframe with equal or less columns.
    """
    if not alwayskeep:
        alwayskeep = []
    if isinstance(alwayskeep, str):
        alwayskeep = [alwayskeep]
    if not isinstance(alwayskeep, list):
        raise TypeError("alwayskeep must be a list")
    if not isinstance(dframe, pd.DataFrame):
        raise TypeError("dropconstants() needs a dataframe")
    if dframe.empty:
        return dframe

    columnstodelete = []
    for col in set(dframe.columns) - set(alwayskeep):
        if len(dframe[col].unique()) == 1:
            columnstodelete.append(col)
    if columnstodelete:
        logging.info("Deleting constant columns %s", str(columnstodelete))
    return dframe.drop(columnstodelete, axis=1)


def df2ecl(
    grid_df: pd.DataFrame,
    keywords: Union[str, List[str]],
    eclfiles: Optional[EclFiles] = None,
    dtype: Optional[Type] = None,
    filename: Optional[str] = None,
    nocomments: bool = False,
) -> str:
    """
    Write an include file with grid data keyword, like PERMX, PORO,
    FIPNUM etc, for the GRID section of the Eclipse deck.

    Output (returned as string and optionally written to file) will then
    contain f.ex::

        PERMX
           3.3 4.1 500.1 8543.0 1223.0 5022.0
           411.455 4433.9
        /

    if the grid contains 8 cells (inactive and active).

    Args:
        grid_df: Dataframe with the keyword for which
            we want to export data, and also the a column with GLOBAL_INDEX.
            Without GLOBAL_INDEX, the output will likely be invalid.
            The grid can contain both active and inactive cells.
        keywords: The keyword(s) to export, with one
            value for every cell.
        eclfiles: If provided, the total cell count for the grid
            will be requested from this object. If not, it will be *guessed*
            from the maximum number of GLOBAL_INDEX, which can be under-estimated
            in the corner-case that the last cells are inactive.
        dtype: If provided, the columns which are
            outputted are converted to int or float. Dataframe columns
            read from CSV files easily gets the wrong type, while Eclipse
            might require some data to be strictly integer.
        filename: If provided, the string produced will also to be
            written to this filename.
        nocomments: Set to True to avoid any comments being written. Defaults
            to False.
    """
    if isinstance(keywords, str):
        keywords = [keywords]

    if isinstance(dtype, str):
        if dtype.startswith("int"):
            dtype = int
        elif dtype.startswith("float"):
            dtype = float
        else:
            raise ValueError(f"Wrong dtype argument {dtype}")

    # Figure out the total number of cells for which we need to export data for:
    global_size = None
    active_cells = None
    if eclfiles is not None:
        if eclfiles.get_egrid() is not None:
            global_size = eclfiles.get_egrid().get_global_size()
            active_cells = eclfiles.get_egrid().getNumActive()

    if "GLOBAL_INDEX" not in grid_df:
        logger.warning(
            (
                "Global index not found in grid dataframe. "
                "Assumes all cells are active"
            )
        )
        # Drop NaN rows for columns to be used (triggered by stacked
        # dates and no global index, unlikely)
        # Also copy dataframe to avoid side-effects on incoming data.
        grid_df = grid_df.dropna(
            axis="rows", subset=[keyword for keyword in keywords if keyword in grid_df]
        )
        grid_df["GLOBAL_INDEX"] = grid_df.index

    if global_size is None:
        global_size = int(grid_df["GLOBAL_INDEX"].max() + 1)
        active_cells = len(grid_df[grid_df.index >= 0])
        logger.warning("Global grid size estimated to %s", str(global_size))

    ecl2df_header = (
        "Output file printed by "
        + "ecl2df.grid "
        + __version__
        + "\n"
        + " at "
        + str(datetime.datetime.now())
    )

    string = ""
    if not nocomments:
        string += common.comment_formatter(ecl2df_header)
    string += "\n"

    # If we have NaNs in the dataframe, we will be more careful (costs memory)
    if grid_df.isna().any().any():
        grid_df = grid_df.dropna(
            axis="rows", subset=[keyword for keyword in keywords if keyword in grid_df]
        )

    for keyword in keywords:
        if keyword not in grid_df.columns:
            raise ValueError(f"Keyword {keyword} not found in grid dataframe")
        vector = np.zeros(global_size)
        vector[grid_df["GLOBAL_INDEX"].astype(int).values] = grid_df[keyword]
        if dtype == int:
            vector = vector.astype(int)
        if dtype == float:
            vector = vector.astype(float)
        if len(vector) != global_size:
            logger.warning(
                (
                    "Mismatch between dumped vector length "
                    "%d from df2ecl and assumed grid size %d"
                ),
                len(vector),
                global_size,
            )
            logger.warning("Data will be dumped, but may error in simulator")
        strvector = "  ".join([str(x) for x in vector])
        strvector = common.runlength_eclcompress(strvector)

        string += keyword + "\n"
        indent = " " * 5
        string += "\n".join(
            textwrap.wrap(
                strvector, initial_indent=indent, subsequent_indent=indent, width=70
            )
        )
        string += "\n/"
        if not nocomments:
            string += (
                f" -- {keyword}: {active_cells} active cells, "
                f"{global_size} total cell count\n"
            )
        string += "\n"

    if filename is not None:
        Path(filename).parent.mkdir(parents=True, exist_ok=True)
        Path(filename).write_text(string, encoding="utf-8")
    return string


def grid_main(args) -> None:
    """This is the command line API"""
    logger = getLogger_ecl2csv(  # pylint: disable=redefined-outer-name
        __name__, vars(args)
    )
    eclfiles = EclFiles(args.DATAFILE)
    grid_df = df(
        eclfiles,
        vectors=args.vectors,
        rstdates=args.rstdates,
        dropconstants=args.dropconstants,
        stackdates=args.stackdates,
    )
    if args.arrow:
        grid_df = _df2pyarrow(grid_df)
    common.write_dframe_stdout_file(
        grid_df, args.output, index=False, caller_logger=logger
    )



"""
Support module for inferring EQLDIMS and TABDIMS from incomplete
Eclipse 100 decks (typically single include-files)
"""

import logging
from typing import Dict, Optional, Union

try:
    import opm.io
except ImportError:
    # Let parts of ecl2df work without OPM:
    pass

from ecl2df import EclFiles

logger = logging.getLogger(__name__)

# Constants to use for pointing to positions in the xxxDIMS keyword
DIMS_POS: Dict[str, int] = dict(NTPVT=1, NTSFUN=0, NTEQUL=0)


def guess_dim(deckstring: str, dimkeyword: str, dimitem: int = 0) -> int:
    """Guess the correct dimension count for an incoming deck (string)

    The incoming deck must in string form, if not, extra data is most
    likely already removed by the opm.io parser. TABDIMS or EQLDIMS
    must not be present

    This function will inject TABDIMS or EQLDIMS into it and reparse it in a
    stricter mode, to detect the correct table dimensionality

    Arguments:
        deck: String containing an Eclipse deck or only a few Eclipse keywords
        dimkeyword: Either TABDIMS or EQLDIMS
        dimitem: The element number in TABDIMS/EQLDIMS to modify
    Returns:
        The lowest number for which stricter opm.io parsing succeeds

    """

    if dimkeyword not in ["TABDIMS", "EQLDIMS"]:
        raise ValueError("Only supports TABDIMS and EQLDIMS")
    if dimkeyword == "TABDIMS":
        if dimitem not in [0, 1]:
            raise ValueError("Only support item 0 and 1 in TABDIMS")
    if dimkeyword == "EQLDIMS":
        if dimitem not in [0]:
            raise ValueError("Only item 0 in EQLDIMS can be estimated")

    # A less than ecl2df-standard permissive opm.io, when using
    # this one opm.io will fail if there are extra records
    # in tables (if NTSFUN in TABDIMS is wrong f.ex):
    opmioparser_recovery_fail_extra_records = [
        ("PARSE_INVALID_KEYWORD_COMBINATION", opm.io.action.ignore),
        ("PARSE_MISSING_INCLUDE", opm.io.action.ignore),
        ("PARSE_MISSING_SECTIONS", opm.io.action.ignore),
        ("PARSE_RANDOM_TEXT", opm.io.action.ignore),
        ("PARSE_UNKNOWN_KEYWORD", opm.io.action.ignore),
        ("SUMMARY_UNKNOWN_GROUP", opm.io.action.ignore),
        ("UNSUPPORTED_*", opm.io.action.ignore),
    ]

    max_guess = 640  # This ought to be enough for everybody
    dimcountguess = 0
    for dimcountguess in range(1, max_guess + 1):
        deck_candidate = inject_dimcount(
            deckstring, dimkeyword, dimitem, dimcountguess, nowarn=True
        )
        try:
            EclFiles.str2deck(
                deck_candidate,
                parsecontext=opm.io.ParseContext(
                    opmioparser_recovery_fail_extra_records
                ),
            )
            # If we succeed, then the dimcountguess was correct
            break
        except (ValueError, RuntimeError):
            # ValueError in opm-common <= 2020.04
            # RuntimeError in opm-common >= 2020.10
            # Typically we get the error PARSE_EXTRA_RECORDS because we did not guess
            # high enough dimnumcount
            continue
            # If we get here, try another dimnumcount
    if dimcountguess == max_guess:
        logger.warning(
            "Unable to guess dim count for %s, or larger than %d", dimkeyword, max_guess
        )
    logger.info("Guessed dimension count count for %s to %d", dimkeyword, dimcountguess)
    return dimcountguess


def inject_dimcount(
    deckstr: str, dimkeyword: str, dimitem: int, dimvalue: int, nowarn: bool = False
) -> str:
    """Insert a TABDIMS with NTSFUN into a deck

    This is simple string manipulation, not opm.io
    deck manipulation (which might be possible to do).

    This function is to be wrapped by inject_xxxdims_ntxxx()

    Arguments:
        deckstr: A string containing a partial deck (f.ex only
            the SWOF keyword).
        dimkeyword: Either TABDIMS or EQLDIMS
        dimitem: Item 0 (NTSSFUN) or 1 (NTPVT) of TABDIMS, only 0 for EQLDIMS.
        dimvalue: The NTSFUN/NTPVT/NTEQUIL number to use
            (this function does not care if it is correct or not but
            it must be larger than zero)
        nowarn: By default it will warn if this function
            is run on a deckstr with TABDIMS/EQLDIMS present. Mute this if True.
    Returns:
        New deck with TABDIMS/EQLDIMS prepended.
    """
    assert dimvalue > 0, "dimvalue must be larger than zero"
    if dimkeyword not in ["TABDIMS", "EQLDIMS"]:
        raise ValueError("Only supports TABDIMS and EQLDIMS")
    if dimkeyword == "TABDIMS":
        if dimitem not in [0, 1]:
            raise ValueError("Only support item 0 and 1 in TABDIMS")
    if dimkeyword == "EQLDIMS":
        if dimitem not in [0]:
            raise ValueError("Only item 0 in EQLDIMS can be injected")

    if dimkeyword in deckstr:
        if not nowarn:
            logger.warning(
                "Not inserting %s in a deck where already exists", dimkeyword
            )
        return deckstr
    return (
        dimkeyword
        + "\n "
        + int(dimitem) * "1* "
        + str(dimvalue)
        + " /\n\n"
        + str(deckstr)
    )


def inject_xxxdims_ntxxx(
    xxxdims: str,
    ntxxx_name: str,
    deck: Union[str, "opm.libopmcommon_python.Deck"],
    ntxxx_value: Optional[int] = None,
) -> "opm.libopmcommon_python.Deck":
    """Ensures TABDIMS/EQLDIMS is present in a deck.

    If ntxxx_value=None and ntxxx_name not in the deck, ntxxx_name will
    be inferred through trial-and-error parsing of the deck, and then injected
    into the deck.

    Args:
        xxxdims: TABDIMS or EQLDIMS
        ntxxx_name: NTPVT, NTEQUL or NTSFUN
        deck: A data deck. If ntxxx_name is to be
            estimated this *must* be a string and not a fully parsed deck.
        npxxx_value: Supply this if ntxxx_name is known, but not present in the
            deck, this will override any guessing. If the deck already
            contains XXXDIMS, this will be ignored.

    Returns:
        opm.io Deck object
    """
    assert xxxdims in ["TABDIMS", "EQLDIMS"]
    assert ntxxx_name in ["NTPVT", "NTEQUL", "NTSFUN"]

    if xxxdims in deck and ntxxx_value is None:
        # Then we have nothing to do, but ensure we parse a potential string to a deck
        if isinstance(deck, str):
            deck = EclFiles.str2deck(deck)
        return deck

    if xxxdims in deck and ntxxx_value is not None:
        logger.warning(
            "Ignoring %s argument, it is already in the deck", str(ntxxx_name)
        )
        return deck

    if not isinstance(deck, str):
        # The deck must be converted to a string deck in order
        # to estimate dimensions.
        deck = str(deck)

    # Estimate if ntxxx_value is not provided:
    if ntxxx_value is None:
        ntxxx_estimate = guess_dim(deck, xxxdims, DIMS_POS[ntxxx_name])
        logger.warning("Estimated %s=%s", ntxxx_name, str(ntxxx_estimate))
    else:
        ntxxx_estimate = ntxxx_value

    augmented_strdeck = inject_dimcount(
        str(deck), xxxdims, DIMS_POS[ntxxx_name], ntxxx_estimate, nowarn=True
    )
    # Overwrite the deck object
    deck = EclFiles.str2deck(augmented_strdeck)

    return deck



from pathlib import Path

import pandas as pd
import pytest

from ecl2df import wellconnstatus
from ecl2df.eclfiles import EclFiles

try:
    import opm  # noqa

    HAVE_OPM = True
except ImportError:
    HAVE_OPM = False

TESTDIR = Path(__file__).absolute().parent
REEK = str(TESTDIR / "data/reek/eclipse/model/2_R001_REEK-0.DATA")
EIGHTCELLS = str(TESTDIR / "data/eightcells/EIGHTCELLS.DATA")


def test_reek_dataset():
    """Test Reek dataset. It contains no CPI data and should return
    an empty dataframe.
    """
    eclfiles = EclFiles(REEK)
    wellconnstatus_df = wellconnstatus.df(eclfiles)
    assert wellconnstatus_df.empty


def test_eightcells_dataset():
    """Test the Eightcells dataset which has CPI data"""
    eclfiles = EclFiles(EIGHTCELLS)
    wellconnstatus_df = wellconnstatus.df(eclfiles)
    expected_dframe = pd.DataFrame(
        [
            {
                "DATE": "2000-01-02",
                "WELL": "OP1",
                "I": "1",
                "J": "1",
                "K": "1",
                "OP/SH": "OPEN",
            }
        ],
    )
    expected_dframe["DATE"] = pd.to_datetime(expected_dframe["DATE"])
    pd.testing.assert_frame_equal(wellconnstatus_df, expected_dframe, check_dtype=False)


@pytest.mark.parametrize(
    "smry, expected_wellconnstatus",
    [
        # Simple Example with one well that is opened.
        # Summary vectors on wrong format is ignored
        (
            pd.DataFrame(
                [
                    ["2000-01-01", 0, 0, 0],
                    ["2000-01-02", 1.1, 1, 1000],
                ],
                columns=["DATE", "CPI:OP1:1,1,1", "CPI:OP1:123", "FOPT"],
            ),
            pd.DataFrame(
                [
                    ["2000-01-02", "OP1", "1", "1", "1", "OPEN"],
                ],
                columns=["DATE", "WELL", "I", "J", "K", "OP/SH"],
            ),
        ),
        # Two connections, and only one is never opened which gives no output
        # The other is opened from first date and then closed
        (
            pd.DataFrame(
                [
                    ["2000-01-01", 0, 1],
                    ["2000-01-02", 0, 0],
                ],
                columns=["DATE", "CPI:OP1:1,1,1", "CPI:OP1:1,1,2"],
            ),
            pd.DataFrame(
                [
                    ["2000-01-01", "OP1", "1", "1", "2", "OPEN"],
                    ["2000-01-02", "OP1", "1", "1", "2", "SHUT"],
                ],
                columns=["DATE", "WELL", "I", "J", "K", "OP/SH"],
            ),
        ),
        # Two wells. Dates containing hours
        (
            pd.DataFrame(
                [
                    ["2000-01-01", 1, 0],
                    ["2000-01-02 12:00:00", 1, 1],
                    ["2000-01-02", 0, 1],
                ],
                columns=["DATE", "CPI:OP1:1,1,1", "CPI:OP2:1,1,1"],
            ),
            pd.DataFrame(
                [
                    ["2000-01-01", "OP1", "1", "1", "1", "OPEN"],
                    ["2000-01-02", "OP1", "1", "1", "1", "SHUT"],
                    ["2000-01-02 12:00:00", "OP2", "1", "1", "1", "OPEN"],
                ],
                columns=["DATE", "WELL", "I", "J", "K", "OP/SH"],
            ),
        ),
    ],
)
def test_extract_status_changes(smry, expected_wellconnstatus):
    """Testing that the extract_status_changes function is working
    correctly with various summary input
    """
    smry["DATE"] = pd.to_datetime(smry["DATE"])
    smry.set_index("DATE", inplace=True)
    expected_wellconnstatus["DATE"] = pd.to_datetime(expected_wellconnstatus["DATE"])

    pd.testing.assert_frame_equal(
        wellconnstatus._extract_status_changes(smry),
        expected_wellconnstatus,
        check_dtype=False,
    )



"""Converter module for Eclipse RFT output files to Pandas Dataframes

If MULTISEG wells are found, the segment data associated to
a connection is merged onto the same row as additional columns,
assuming connections do not point to more than one segment.

If ICD segments are detected (recognized as branches only containing
one segment), they are merged into the same row that already contains
connection data (CONxxxxx) and its segment data (now giving
information for the conditions in the tubing).

The columns representing SEGxxxxx data on ICD segments are renamed
by adding the prefix ``ICD_``
"""

import argparse
import collections
import datetime
import logging
from typing import Any, Dict, Iterable, Optional, Set

import numpy as np
import pandas as pd
from ecl.eclfile import EclFile

from ecl2df import getLogger_ecl2csv

from .common import merge_zones, write_dframe_stdout_file
from .eclfiles import EclFiles
from .gruptree import tree_from_dict

logger: logging.Logger = logging.getLogger(__name__)

# In debug mode, these columns will be exported to three csv files.
CON_TOPOLOGY_COLS: Set = {"CONIDX", "CONBRNO", "CONSEGNO", "CONNXT", "DEPTH"}
SEG_TOPOLOGY_COLS: Set = {
    "SEGIDX",
    "SEGIDX_upstream",
    "SEGBRNO",
    "SEGBRNO_upstream",
    "SEGNXT",
    "SEGNXT_upstream",
    "JUNCTION",
    "JUNCTION_downstream",
    "LONELYSEG",
    "LEAF",
    "SEGDEPTH",
}
ICD_TOPOLOGY_COLS: Set = {
    "ICD_SEGBRNO_upstream",
    "ICD_SEGIDX_upstream",
    "ICD_LEAF",
    "ICD_JUNCTION",
    "ICD_LONELYSEG",
    "ICD_JUNCTION_downstream",
    "ICD_SEGBRNO",
    "ICD_SEGNXT",
    "ICD_SEGIDX",
    "ICD_SEGDEPTH",
}


def _rftrecords2df(rftfile: EclFile) -> pd.DataFrame:
    """Construct a dataframe just for navigation on the RFT records,
    from the attribute 'headers' in EclFile object constructed from the
    binary RFT file

    The dataframe will consist of the columns and with example data:
        timeindex, recordidx, recordname, recordlength, recordtype
        0, 0, TIME, 1, REAL
        0, 1, DATE, 3, INTE
        0, 2, WELLETC, 16,  CHAR
        ...
        1, 30, TIME, 1, REAL
        1, 31, DATE, 3, INTE
        ....
        3, 89, SWAT, 14, REAL

    where the column recordidx refers to the RFT file such that
        rftfile[89] = EclKW(size=14, name="SWAT", ...)

    Args:
        rftfile (EclFile)
    """
    nav_df = pd.DataFrame(rftfile.headers)
    nav_df.columns = ["recordname", "recordlength", "recordtype"]
    nav_df["timeindex"] = np.nan
    # the TIME record (in recordname) signifies that the forthcoming records
    # belong to  this TIME value, and we make a new column in the header data that
    # tells us the row number for the associated TIME record
    nav_df.loc[nav_df["recordname"] == "TIME", "timeindex"] = nav_df[
        nav_df["recordname"] == "TIME"
    ].index
    nav_df.fillna(
        method="ffill", inplace=True
    )  # forward fill (because any record is associated to the previous TIME record)
    nav_df["timeindex"] = nav_df["timeindex"].astype(int)
    logger.info(
        "Located %s RFT headers at %s distinct dates",
        str(len(nav_df)),
        str(len(nav_df["timeindex"].unique())),
    )
    nav_df.index.name = "recordidx"
    return nav_df.reset_index()


def rftrecords(rftfile: EclFile) -> Iterable[Dict[str, Any]]:
    """Generator for looping over RFT records in a EclFile object.

    Each returned RFT record is represented as a dict with the keys:
        headers: pd.DataFrame, indexed by recordname

    Args:
        EclFile made from a binary RFT file.
    """
    navigation_frame = _rftrecords2df(rftfile)
    for timeindex, headers in navigation_frame.groupby("timeindex"):
        headers = headers.set_index("recordname")
        rftrecord = {}
        rftrecord["headers"] = headers
        # All rows in nav_record_df represents  the data in the current
        # RFT record
        dateidx = int(headers.loc["DATE"]["recordidx"])
        rftrecord["date"] = datetime.date(
            rftfile[dateidx][2], rftfile[dateidx][1], rftfile[dateidx][0]
        )
        rftrecord["wellname"] = rftfile[int(headers.loc["WELLETC"]["recordidx"])][
            1
        ].strip()

        rftrecord["wellmodel"] = rftfile[int(headers.loc["WELLETC"]["recordidx"])][
            6
        ].strip()
        # wellmodel is either "STANDARD" or "MULTISEG"

        rftrecord["timeindex"] = timeindex
        yield rftrecord


def get_con_seg_data(
    rftrecord: Dict[str, Any], rftfile: EclFile, datatype: str
) -> pd.DataFrame:
    """
    Build a dataframe of CON* or SEG* data for a specific RFT record,
    that is for one well at one date.

    Dataframe will for datatype=="CON" look like::

      DEPTH, SWAT, CONKH, CONIDX, ..
      2300,  0.3, 3000, 1
      2310, 0.2, 1231, 2

    and number of rows will equal the number of connected cells (COMPDAT lines)

    If it is for SEG data, all columns are prefixed by SEG
    and row count will be the number of segments defined in WELSEGS

    Args:
        rftrecord: Data for one RFT record, provided by rftrecords()
        rftfile:
        datatype: Either "CON" or "SEG"
    """
    if datatype not in ["CON", "SEG"]:
        raise ValueError("datatype must equal CON or SEG")

    headers = rftrecord["headers"]
    if datatype == "CON":
        rft_row_count = headers.loc["DEPTH"]["recordlength"]
    else:
        rft_row_count = headers.loc["SEGDEPTH"]["recordlength"]

    data_headers = headers[headers.recordlength == rft_row_count].reset_index()[
        ["recordidx", "recordname"]
    ]
    # If CON type, ensure, no SEG data included, and vice versa
    if datatype == "CON":
        data_headers = data_headers[~data_headers["recordname"].str.startswith("SEG")]
    else:
        data_headers = data_headers[data_headers["recordname"].str.startswith("SEG")]

    data = pd.DataFrame(
        {
            row["recordname"]: list(rftfile[row["recordidx"]])
            for _, row in data_headers.iterrows()
        }
    )

    # Ensure integer headers are integers:
    integer_columns = headers[headers["recordtype"] == "INTE"].index.values
    for col in (set(integer_columns) - {"DATE"}).intersection(
        set(data_headers["recordname"].values)
    ):
        data[col] = data[col].astype(int)
    data[datatype + "IDX"] = data.index + 1  # Add an index that starts with 1
    return data


def count_wellbranches(seg_data: pd.DataFrame) -> int:
    """From a segment dataframe, coming from get_con_seg_data(..., "SEG")
    determine the number of well branche.

    ICD segments must be split out first using split_seg_icd(), otherwise
    results are not reliable.

    Args:
        pd.DataFrame, with at least the columns SEGIDX, SEGNXT and SEGBRNO
    """
    if "LEAF" not in seg_data:
        seg_data = process_seg_topology(seg_data)

    branchcount = len(
        seg_data[~seg_data["LEAF"] | seg_data["JUNCTION_downstream"]]["SEGBRNO"]
        .dropna()
        .unique()
    )

    # logger.debug("Branches found: %d", count_wellbranches(merged))
    return max(1, branchcount)


def process_seg_topology(seg_data: pd.DataFrame) -> pd.DataFrame:
    """Determine and process the segment topology.

    The topology of the well segments are determined by the SEGNXT column in
    incoming dataframe, which corresponds to SEGIDX.

    SEGNXT points to the next segment downstream in a production well (and
    injector wells are treated as production here). Downstream is thus *upwards*
    in space, towards the sea.

    The last segment in a non-icd well gets the type TUBING.

    Args:
        seg_data: Segment structure defined as a table with at least
            the columns SEGIDX, SEGNXT

    Returns:
        Augmented dataframe, extra columns and perhaps extra rows.
    """
    if not {"SEGIDX", "SEGNXT"}.issubset(set(seg_data.columns)):
        raise ValueError("Insufficient topology columns in dataframe")

    seg_data = seg_data.sort_values("SEGIDX")
    # For the first segment, None is allowed as SEGNXT, which excludes
    # int as a  Pandas type. Convert to 0 for the moment
    seg_data["SEGNXT"] = seg_data["SEGNXT"].fillna(value=0).astype(int)

    # Outer merge first to add the upstream segment information to every row.
    merged = pd.merge(
        seg_data,
        seg_data,
        how="left",
        left_on="SEGIDX",
        right_on="SEGNXT",
        suffixes=("", "_upstream"),
    )
    del merged["SEGNXT_upstream"]
    # Later this might be changed to use the Pandas nullable integer type, but
    # using zero for NaN works in this context.
    merged["SEGIDX_upstream"] = merged["SEGIDX_upstream"].fillna(value=0).astype(int)
    merged["SEGBRNO_upstream"] = merged["SEGBRNO_upstream"].fillna(value=0).astype(int)

    # Now we can determine leaf segments by those with no extra information, since
    # we did an outer merge:
    merged["LEAF"] = merged["SEGIDX_upstream"].replace(0, np.nan).isnull()

    # Flag segments that have multiple upstream segments as junctions
    merged["JUNCTION"] = merged["SEGIDX"].duplicated(keep=False)

    # Determine if a segment is alone on its own branch
    merged["LONELYSEG"] = ~merged["SEGBRNO"].duplicated(keep=False)

    # We also want to flag the segment that is upstream a junction,
    merged["JUNCTION_downstream"] = False
    merged.loc[
        merged[merged["JUNCTION"]]["SEGIDX_upstream"], "JUNCTION_downstream"
    ] = True

    return merged


def seg2dicttree(seg_data: pd.DataFrame) -> dict:
    """Generate a nested dictionary representing the
    well through its segment topology

    Args:
        seg_data: topology determined by SEGIDX
            and SEGIDX_upstream

    Returns:
        Nested dict,
    """
    if seg_data.empty:
        return {}

    if "LEAF" not in seg_data:
        seg_data = process_seg_topology(seg_data)
    subtrees: dict = collections.defaultdict(dict)
    edges = []  # List of tuples
    for _, row in seg_data.iterrows():
        if "SEGIDX_upstream" in row and row["SEGIDX_upstream"] > 0:
            edges.append((row["SEGIDX_upstream"], row["SEGIDX"]))
    if not edges:
        return {seg_data["SEGIDX"].values[0]: {}}
    for child, parent in edges:
        subtrees[parent][child] = subtrees[child]

    children, parents = zip(*edges)
    roots = set(parents).difference(children)
    trees = []
    trees.append({root: subtrees[root] for root in roots})
    return trees[0]


def pretty_print_well(seg_data: pd.DataFrame) -> str:
    """Return a multiline string with the segment structure
    pretty printed as an ASCII tree.

    Args:
        seg_data: Segment dataframe

    Returns:
        Multiline string
    """
    dicttree = seg2dicttree(seg_data)
    return str(tree_from_dict(dicttree))


def split_seg_icd(seg_data: pd.DataFrame) -> pd.DataFrame:
    """Split a segment dataframe into a dataframe
    with non-ICD segments and one with.

    The segment properties (data) are merged into the downstream segments
    dataset, with the SEG prefixed switched to ICD.

    Returns:
        Dataframe with the ICD segments only. Empty if no ICDs found.
        and wider.
    """

    # Ensure we have some topology data present:
    if "LEAF" not in seg_data:
        seg_data = process_seg_topology(seg_data)

    icd_present = seg_data["SEGBRNO"].max() > count_wellbranches(seg_data)

    if not icd_present:
        return (seg_data, pd.DataFrame())

    # ICD segments are those where:
    #  * Leaf segments (connected reservoir / con_data row)
    #  # * Downstream segment is not a junction (because it must
    #    be connected to a tubing segment, which is again connected
    #    to a junction) (this might be too strict?)
    #    STOP: Cannot use this criteria, because junctions  due to ICDs
    #    are legit.
    #  * The segment must be on a branch with only one segment
    icd_seg_indices = seg_data[seg_data["LEAF"] & seg_data["LONELYSEG"]].index.values
    non_icd_seg_indices = seg_data[
        ~(seg_data["LEAF"] & seg_data["LONELYSEG"])
    ].index.values

    icd_seg_data = seg_data.reindex(icd_seg_indices)
    seg_data = seg_data.reindex(non_icd_seg_indices)

    icd_seg_data.columns = ["ICD_" + x for x in icd_seg_data.columns]

    logger.debug(
        "Found %d ICD segments, indices %s",
        len(icd_seg_data),
        str(icd_seg_data["ICD_SEGIDX"].values),
    )

    return (seg_data, icd_seg_data)


def merge_icd_seg_conseg(
    con_data: pd.DataFrame,
    seg_data: Optional[pd.DataFrame] = None,
    icd_data: Optional[pd.DataFrame] = None,
) -> pd.DataFrame:
    """
    Merge ICD segments to the CONxxxxx data. We will be
    connection-centric in the outputted rows, that is
    one row pr. connection. If the setup is with more
    than one segment pr. connection (e.g. reservoir
    cell), then we would have to be smarter. Either
    averaging the properties, or be segment-centric in
    the output.

    Petrel happily puts many ICD segments to the same
    connection. This setup is a bug, with partially
    unknown effects when simulated in Eclipse. Should we
    warn the user??

    Args:
        con_data: Connection data (CONxxxx columns). One
            row pr. reservoir connection
        seg_data: Segment data, SEGxxxxx cols, one row pr. segment, and
            each segment should correspond to one ICD or to one reservoir connection.
            Supply empty or None if no segments found.
        icd_data: ICD data, ICD_SEGxxxx columns. One row pr
            ICD segment. One-to-one correspondence to con_data through ICD_SEGBRNO
            and CONBRNO required. Can be empty or None if no ICD present.
    """
    if seg_data is None:
        seg_data = pd.DataFrame()
    if icd_data is None:
        icd_data = pd.DataFrame()

    if logger.level <= logging.DEBUG:
        logger.debug("Writing connection data to con.csv")
        con_data[CON_TOPOLOGY_COLS.intersection(con_data.columns)].to_csv(
            "con.csv", index=False
        )
        logger.debug("Writing segment data to seg.csv")
        seg_data[SEG_TOPOLOGY_COLS.intersection(seg_data.columns)].to_csv(
            "seg.csv", index=False
        )
        logger.debug("Writing ICD data to icd.csv")
        icd_data[ICD_TOPOLOGY_COLS.intersection(icd_data.columns)].to_csv(
            "icd.csv", index=False
        )

    data = pd.DataFrame()
    if not icd_data.empty:
        # Merge ICD_* columns onto the dataframe representing reservoir
        # connections.
        data = pd.merge(con_data, icd_data, left_on="CONSEGNO", right_on="ICD_SEGIDX")

        # Merge SEGxxxxx to the dataframe with icd's and reservoir connections.
        assert not seg_data.empty

        data = pd.merge(
            data, seg_data, how="left", left_on="ICD_SEGNXT", right_on="SEGIDX"
        )

        # The merge has potentially included extra rows due to junctions.
        # After ICD merge, we can require that SEGIDX_upstream equals CONSEGNO
        data = data[data["CONSEGNO"] == data["SEGIDX_upstream"]]

        # Now the data for a reservoir connection, its ICD segment and the tubing
        #  segment is on the same row in the dataframe.

        # Gather connections that are not associated to ICDs:
        no_icd_con_segments = set(con_data["CONSEGNO"]) - set(icd_data["ICD_SEGIDX"])
        con_data_no_icd = (
            con_data.set_index("CONSEGNO").loc[no_icd_con_segments].reset_index()
        )
    else:
        con_data_no_icd = con_data

    if not seg_data.empty:
        data = pd.concat(
            [
                data,
                pd.merge(
                    con_data_no_icd, seg_data, left_on="CONSEGNO", right_on="SEGIDX"
                ),
            ],
            sort=False,
        )
    else:
        # Non-multisegment wells have only reservoir connection data.
        return con_data
    return data


def add_extras(dframe: pd.DataFrame, inplace: bool = True) -> pd.DataFrame:
    """Add extra nice-to-have columns to the dataframe
    with tubing, icd-segments, and reservoir  connections matched
    on rows

    Args:
        dframe: Dataframe typically obtained from merge_icd_seg()
        inplace: Set to False if the original should not be modified.

    Returns:
        The (possibly) augmented incoming dataframe.
    """
    if not inplace:
        dframe = dframe.copy()

    if "CONPRES" in dframe and "SEGPRES" in dframe:
        dframe["COMPLETION_DP"] = 0
        nonzero_pres = (dframe["CONPRES"] > 0) & (dframe["SEGPRES"] > 0)
        dframe.loc[nonzero_pres, "COMPLETION_DP"] = (
            dframe.loc[nonzero_pres, "CONPRES"] - dframe.loc[nonzero_pres, "SEGPRES"]
        )

    if not dframe.empty:
        dframe["DRAWDOWN"] = 0  # Set a default so that the column always exists

    if "CONPRES" in dframe and "PRESSURE" in dframe:
        nonzero_conpres = dframe["CONPRES"] > 0
        dframe.loc[nonzero_conpres, "DRAWDOWN"] = (
            dframe.loc[nonzero_conpres, "PRESSURE"]
            - dframe.loc[nonzero_conpres, "CONPRES"]
        )

    if "PRESSURE" in dframe:
        dframe["CONBPRES"] = dframe["PRESSURE"]  # Just an alias
    if "CONLENEN" in dframe and "CONLENST" in dframe:
        dframe["CONMD"] = 0.5 * (dframe["CONLENST"] + dframe["CONLENEN"])
        dframe["CONLENTH"] = dframe["CONLENEN"] - dframe["CONLENST"]

    if "CONORAT" in dframe and "CONLENTH" in dframe:
        dframe["CONORATS"] = dframe["CONORAT"] / dframe["CONLENTH"]
    if "CONWRAT" in dframe and "CONLENTH" in dframe:
        dframe["CONWRATS"] = dframe["CONWRAT"] / dframe["CONLENTH"]
    if "CONGRAT" in dframe and "CONLENTH" in dframe:
        dframe["CONGRATS"] = dframe["CONGRAT"] / dframe["CONLENTH"]

    return dframe


def df(
    eclfiles: EclFiles, wellname: Optional[str] = None, date: Optional[str] = None
) -> pd.DataFrame:
    """Loop over an RFT file and construct a dataframe representation
    of the data, ordered by well and date.

    Args:
        eclfiles: Object used to locate the RFT file
        wellname: If provided, only wells matching this string exactly
            will be included
        date: If provided, all other dates will be ignored. YYYY-MM-DD.
    """
    rftfile = eclfiles.get_rftfile()

    rftdata = []
    for rftrecord in rftrecords(rftfile):
        if wellname is not None and rftrecord["wellname"] != wellname:
            continue
        if date is not None and str(rftrecord["date"]) != date:
            continue

        logger.info(
            "Extracting %s well %s at %s, record index: %s",
            rftrecord["wellmodel"],
            str.rjust(rftrecord["wellname"], 8),
            rftrecord["date"],
            rftrecord["timeindex"],
        )

        headers = rftrecord["headers"]

        if "DEPTH" not in headers.index:
            logger.debug(
                "Well %s has no data to extract at %s",
                str(rftrecord["wellname"]),
                str(rftrecord["date"]),
            )
            continue

        con_data = get_con_seg_data(rftrecord, rftfile, "CON")

        # Process multisegment data (not necessarily the same number
        # of rows as the connection data). Data for segments
        # that are not associated with a connection will not be
        # included.

        has_seg_data = any(headers.index.str.startswith("SEG"))
        if rftrecord["wellmodel"] == "MULTISEG" and not has_seg_data:
            logger.warning(
                "Well %s is a multisegment well, but has no SEG data",
                rftrecord["wellname"],
            )
            # This should probably never happen (?)

        seg_data = pd.DataFrame()
        icd_data = pd.DataFrame()
        if has_seg_data:
            seg_data = get_con_seg_data(rftrecord, rftfile, "SEG")

            # For each downstream segment, merge in the data for its
            # upstream segment, and determine leaf nodes:
            seg_data = process_seg_topology(seg_data)
            logger.debug(pretty_print_well(seg_data))

            # NB: The enumeration of branches is not necessarily consecutive
            # from SEGBRNO.

            # Now we can test if we have any ICD segments, that is the
            # case if we have any segments that have SEGBRNO higher than
            # the branch count.

            seg_data, icd_data = split_seg_icd(seg_data)

            # Branch counting must be done after ICD's are split out.
            branchcount = count_wellbranches(seg_data)

            logger.info(
                "Found %d branch(es), and %d icd segment(s)", branchcount, len(icd_data)
            )

        con_icd_seg = merge_icd_seg_conseg(con_data, seg_data, icd_data)

        con_icd_seg = add_extras(con_icd_seg, inplace=True)

        con_icd_seg["DATE"] = rftrecord["date"]
        con_icd_seg["WELL"] = rftrecord["wellname"]
        con_icd_seg["WELLMODEL"] = rftrecord["wellmodel"]

        # Delete topology columns
        delete_cols = {col for col in con_icd_seg.columns if col.endswith("stream")}
        delete_cols = delete_cols.union(
            {
                "LEAF",
                "ICD_LEAF",
                "JUNCTION",
                "ICD_JUNCTION",
                "LONELYSEG",
                "ICD_LONELYSEG",
            }
        )
        con_icd_seg = con_icd_seg[set(con_icd_seg.columns) - delete_cols]

        rftdata.append(con_icd_seg)

    rftdata_df = pd.concat(rftdata, ignore_index=True, sort=False)

    # Fill empty cells with zeros. This is to avoid Spotfire
    # interpreting columns with numbers as strings. An alternative
    # solution that keeps NaN would be to add a second row in the
    # output containing the datatype
    rftdata_df.fillna(0, inplace=True)

    # The HOSTGRID data seems often to be empty, check if it is and delete if so:
    if "HOSTGRID" in rftdata_df.columns:
        if len(rftdata_df.HOSTGRID.unique()) == 1:
            if rftdata_df.HOSTGRID.unique()[0].strip() == "":
                del rftdata_df["HOSTGRID"]

    zonemap = eclfiles.get_zonemap()
    if zonemap:
        if "K" in rftdata_df:
            kname = "K"
        else:
            kname = "CONKPOS"
        rftdata_df = merge_zones(rftdata_df, zonemap, kname=kname)

    return rftdata_df


def fill_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Set up sys.argv parsers.

    Arguments:
        parser to fill with arguments
    """
    parser.add_argument(
        "DATAFILE",
        help=(
            "Name of Eclipse DATA file or RFT file. "
            "If DATA file is provided, it will look for"
            " the associated DATA file"
        ),
    )
    parser.add_argument(
        "--wellname", type=str, help="Restrict data to one named well", default=None
    )

    parser.add_argument(
        "--date", type=str, help="Restrict data to one date, YYYY-MM-DD", default=None
    )
    parser.add_argument(
        "-o", "--output", type=str, help="Name of output CSV file.", default="rft.csv"
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose")
    parser.add_argument("--debug", action="store_true", help="Debug mode")
    return parser


def rft_main(args) -> None:
    """Entry-point for module, for command line utility"""
    logger = getLogger_ecl2csv(  # pylint: disable=redefined-outer-name
        __name__, vars(args)
    )
    if args.DATAFILE.endswith(".RFT"):
        # Support the RFT file as an argument also:
        eclfiles = EclFiles(args.DATAFILE.replace(".RFT", "") + ".DATA")
    else:
        eclfiles = EclFiles(args.DATAFILE)
    rft_df = df(eclfiles, wellname=args.wellname, date=args.date)
    if rft_df.empty:
        if args.wellname is not None or args.date is not None:
            logger.warning("No data. Check your well and/or date filter")
        else:
            logger.error("No data found. Bug?")
        return
    write_dframe_stdout_file(rft_df, args.output, index=False, caller_logger=logger)


# Vector  Description
# CONIDX   Index of the connection pr well, starting at 1.
# CONDEPTH Depth at the centre of each connection in the well
# CONLENST Length down the tubing from the BH reference
#          point to the start of the connection
# CONLENEN Length down the tubing from the BH reference point to the
#          far end of the connection
# CONPRES  Pressure in the wellbore at the connection
# CONORAT  Oil production rate of the connection at surface conditions
# CONWRAT  Water production rate of the connection at surface conditions
# CONGRAT  Gas production rate of the connection at surface conditions
# CONOTUB  Oil flow rate through the tubing at the start of the
#          connection at surface conditions
# CONWTUB  Water flow rate through the tubing at the start of the
#          connection at surface conditions
# CONGTUB  Gas flow rate through the tubing at the start of the
#          connection at surface conditions
# CONVTUB  Volumetric flow rate of the mixture at the start of the connection
# CONFAC   Connection transmissibility factor
# CONKH    Connection Kh value
# CONNXT   Number of the neighbouring connection towards the wellhead
# CONSEGNO Segment number containing the connection
# CONBRNO  Branch number containing the connection
# CONIPOS  I location of the connection
# CONJPOS  J location of the connection
# CONKPOS  K location of the connection
# CONBDEPH Depth of the grid block of the connection
# CONBPRES Pressure of the grid block of the connection
#          (Copy of the PRESSURE data)
# CONBSWAT Water saturation of the grid block of the connection
# CONBSGAS Gas saturation of the grid block of the connection
# CONBSOIL Oil saturation of the grid block of the connection
# COMPLETION Completion index of the connection
#
# The above values are taken from the corresponding RFT data.
#
# Vector   Description
# CONMD    Measured depth of the connection
# CONLENTH Length of the connection
# CONORATS Scaled oil production rate at surface conditions
# CONWRATS Scaled water production rate at surface conditions
# CONGRATS Scaled gas production rate at surface conditions
#
#
# Vector   Description
# SEGDEPTH Depth at the far end of each segment
# SEGLENST Length down the tubing from the zero tubing length
#          reference point to the start of the segment
# SEGLELEN Length down the tubing from the zero tubing length
#          reference point to the far end of the segment
# SEGXCORD X-coordinate at the far end of the segment
#          (as entered by the 11th item of the WELSEGS record)
# SEGXCORD Y-coordinate at the far end of the segment
#          (as entered by the 12th item of the WELSEGS record)
# SEGPRES Pressure in the wellbore at the far end of the segment
# SEGORAT Oil flow rate through the segment through its near end
# SEGWRAT Water flow rate through the segment through its near end
# SEGGRAT Gas flow rate through the segment through its near end
# SEGOVEL Free oil phase velocity through the segment
# SEGWVEL Water flow velocity through the segment
# SEGGVEL Free gas phase flow velocity through the segment
# SEGOHF  Free oil phase holdup fraction in the segment
# SEGWHF  Water holdup fraction in the segment
# SEGGHF  Free gas phase holdup fraction in the segment
# SEGBRNO Branch number of the segment
# SEGNXT  Number of the neighbouring segment towards the wellhead
#
# Vector         Description
# SEGMD          Segment measured depth
# SEGLENTH       Segment length
# SEGORATSScaled water flow rate through the segment
# SEGWRATSScaled water flow rate through the segment
# SEGGRATSScaled gas flow rate through the segment
# SEGCORATSummed connection oil flow rate through segment
# SEGCWRATSummed connection water flow rate through segment
# SEGCGRATSummer connection gas flow rate through segment
# SEGCORTSScaled summed connection oil flow rate through segment
# SEGCWRTSScaled summed connection water flow rate through segment
# SEGCGRTSScaled summed connection gas flow rate through segment
